"""This module allows to save and load objects in PNML. The standard
is correctly handled for low-level nets, ie, nets with only `True`
guards and black tokens. However, module `snakes.pnml` was actually
created long before the PNML standard was extended to handle
high-level nets (in particular coloured variants) and so the standard
is not respected in this case. Consequently, in the following, we
abuse the term PNML for the XML produced and read by SNAKES.

@warning: this module will be replaced in a future version of SNAKES,
    so its documentation will be kept minimal.
"""

import xml.dom.minidom
import pickle
import sys, inspect, os, os.path, imp, pkgutil
import snakes, snakes.plugins
from snakes import SnakesError
from snakes.compat import *
if PY3 :
    import ast

try :
    builtins = sys.modules["__builtin__"]
except KeyError :
    builtins = sys.modules["builtins"]

def _snk_import (name) :
    "Properly import a module, including a plugin"
    if name.startswith("snakes.plugins.") :
        return snakes.plugins.load(name, "snakes.nets")
    else :
        return __import__(name, fromlist=["__name__"])

def _snk_modules () :
    "List all SNAKES' modules"
    queue = ["snakes"]
    while len(queue) > 0 :
        modname = queue.pop(0)
        try :
            mod = _snk_import(modname)
        except :
            continue
        yield modname, mod
        importer = pkgutil.ImpImporter(mod.__path__[0])
        for name, ispkg in importer.iter_modules(prefix=mod.__name__ + ".") :
            if ispkg :
                queue.append(name)
            else :
                try :
                    yield name, _snk_import(name)
                except :
                    pass

def _snk_tags () :
    "Lists all PNML tags found in SNAKES"
    for modname, mod in _snk_modules() :
        for clsname, cls in inspect.getmembers(mod, inspect.isclass) :
            if cls.__module__ == modname and "__pnmltag__" in cls.__dict__ :
                yield cls.__pnmltag__, modname, clsname

class _set (object) :
    """Set where items are iterated by order of insertion
    """
    def __init__ (self, elements=[]) :
        """
        >>> _set([4, 5, 1, 2, 4])
        _set([4, 5, 1, 2])
        """
        self._data = {}
        self._last = 0
        for e in elements :
            self.add(e)
    def __repr__ (self) :
        """
        >>> _set([4, 5, 1, 2, 4])
        _set([4, 5, 1, 2])
        """
        return "%s([%s])" % (self.__class__.__name__,
                             ", ".join(repr(x) for x in self))
    def add (self, element) :
        """
        >>> s = _set([4, 5, 1, 2, 4])
        >>> s.add(0)
        >>> s
        _set([4, 5, 1, 2, 0])
        >>> s.add(5)
        >>> s
        _set([4, 5, 1, 2, 0])
        """
        if element not in self._data :
            self._data[element] = self._last
            self._last += 1
    def __contains__ (self, element) :
        """
        >>> 4 in _set([4, 5, 1, 2, 4])
        True
        >>> 0 in _set([4, 5, 1, 2, 4])
        False
        """
        return element in self._data
    def _2nd (self, pair) :
        """
        >>> _set()._2nd((1, 2))
        2
        """
        return pair[1]
    def __iter__ (self) :
        """
        >>> list(_set([4, 5, 1, 2, 4]))
        [4, 5, 1, 2]
        """
        return (k for k, v in sorted(self._data.items(), key=self._2nd))
    def discard (self, element) :
        """
        >>> s = _set([4, 5, 1, 2, 4])
        >>> s.discard(0)
        >>> s.discard(4)
        >>> s
        _set([5, 1, 2])
        """
        if element in self._data :
            del self._data[element]
    def remove (self, element) :
        """
        >>> s = _set([4, 5, 1, 2, 4])
        >>> s.remove(0)
        Traceback (most recent call last):
         ...
        KeyError: ...
        >>> s.remove(4)
        >>> s
        _set([5, 1, 2])
        """
        del self._data[element]
    def copy (self) :
        """
        >>> _set([4, 5, 1, 2, 4]).copy()
        _set([4, 5, 1, 2])
        """
        return self.__class__(self)
    def __iadd__ (self, other) :
        """
        >>> s = _set([4, 5, 1, 2, 4])
        >>> s += range(7)
        >>> s
        _set([4, 5, 1, 2, 0, 3, 6])
        """
        for element in other :
            self.add(element)
        return self
    def __add__ (self, other) :
        """
        >>> _set([4, 5, 1, 2, 4]) + range(7)
        _set([4, 5, 1, 2, 0, 3, 6])
        """
        result = self.copy()
        result += other
        return result
    def __len__ (self) :
        """
        >>> len(_set([4, 5, 1, 2, 4]))
        4
        """
        return len(self._data)

# apidoc skip
class Tree (object) :
    """Abstraction of a PNML tree

    >>> Tree('tag', 'data', Tree('child', None), attr='attribute value')
    <?xml version="1.0" encoding="utf-8"?>
    <pnml>
     <tag attr="attribute value">
      <child/>
      data
     </tag>
    </pnml>
    """
    def __init__ (self, _name, _data, *_children, **_attributes) :
        """Initialize a PNML tree

        >>> Tree('tag', 'data',
        ...      Tree('first_child', None),
        ...      Tree('second_child', None),
        ...      first_attr='attribute value',
        ...      second_attr='another value')
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <tag first_attr="attribute value" second_attr="another value">
          <first_child/>
          <second_child/>
          data
         </tag>
        </pnml>

        Note: parameters names start with a '_' in order to allow for
        using them as attributes.

        @param _name: the name of the tag
        @type _name: `str`
        @param _data: the text held by the tag or `None`
        @type _data: `str` or `None`
        @param _children: children nodes
        @type _children: `Tree`
        @param _attributes: attributes and values of the tag
        @type _attributes: `str`
        """
        self.name = _name
        if _data is not None and _data.strip() == "" :
            _data = None
        self.data = _data
        self.children = list(_children)
        self.attributes = _attributes
    @classmethod
    def _load_tags (_class) :
        if not hasattr(_class, "_tags") :
            _class._tags = {}
            for tag, mod, cls in _snk_tags() :
                if tag not in _class._tags :
                    _class._tags[tag] = (mod, cls)
    def _update_node (self, doc, node) :
        for key, val in self.attributes.items() :
            node.setAttribute(key, val)
        for child in self.children :
            node.appendChild(child._to_dom(doc))
        if self.data is not None :
            node.appendChild(doc.createTextNode(self.data))
    def _to_dom (self, doc) :
        result = doc.createElement(self.name)
        self._update_node(doc, result)
        return result
    def to_pnml (self) :
        """Dumps a PNML tree to an XML string

        >>> print(Tree('tag', 'data', Tree('child', None), attr='value').to_pnml())
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <tag attr="value">
          <child/>
          data
         </tag>
        </pnml>

        @return: the XML string that represents the PNML tree
        @rtype: `str`
        """
        if self.name == "pnml" :
            tree = self
        else :
            tree = self.__class__("pnml", None, self)
        try :
            plugins = _set(self.__plugins__)
        except AttributeError :
            plugins = _set()
        for node in self.nodes() :
            if hasattr(node, "_plugins") :
                plugins += node._plugins
        if len(plugins) > 0 :
            tree.children.insert(0, Tree("snakes", None,
                                         Tree("plugins", None,
                                              Tree.from_obj(tuple(plugins))),
                                         version=snakes.version))
        impl = xml.dom.minidom.getDOMImplementation()
        doc = impl.createDocument(None, "pnml", None)
        node = tree._to_dom(doc)
        tree._update_node(doc, doc.documentElement)
        if len(plugins) > 0 :
            del tree.children[0]
        r = doc.toprettyxml(indent=" ",
                            encoding=snakes.defaultencoding).strip()
        if PY3 :
            return r.decode()
        else :
            return r
    @classmethod
    def from_dom (cls, node) :
        """Load a PNML tree from an XML DOM representation

        >>> src = Tree('object', '42', type='int').to_pnml()
        >>> dom = xml.dom.minidom.parseString(src)
        >>> Tree.from_dom(dom.documentElement)
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <object type="int">
          42
         </object>
        </pnml>

        @param node: the DOM node to load
        @type node: `xml.dom.minidom.Element`
        @return: the loaded PNML tree
        @rtype: `Tree`
        """
        result = cls(node.tagName, node.nodeValue)
        for i in range(node.attributes.length) :
            name = node.attributes.item(i).localName
            result[name] = str(node.getAttribute(name))
        if node.hasChildNodes() :
            for child in node.childNodes :
                if child.nodeType == child.TEXT_NODE :
                    result.add_data(str(child.data))
                elif child.nodeType == child.ELEMENT_NODE :
                    result.add_child(cls.from_dom(child))
        return result
    @classmethod
    def from_pnml (cls, source, plugins=[]) :
        """Load a PNML tree from an XML string representation

        >>> src = Tree('object', '42', type='int').to_pnml()
        >>> Tree.from_pnml(src)
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <object type="int">
          42
         </object>
        </pnml>

        @param source: the XML string to load or an opened file that
            contains it
        @type source: `str` or `file`
        @return: the loaded PNML tree
        @rtype: `Tree`
        """
        try :
            doc = xml.dom.minidom.parse(source)
        except :
            doc = xml.dom.minidom.parseString(source)
        result = cls.from_dom(doc.documentElement)
        plugins = _set(plugins)
        cls._load_tags()
        tag2obj = {}
        trash = []
        for node in result.nodes() :
            node._tag2obj = tag2obj
            if node.has_child("snakes") :
                snk = node.child("snakes")
                trash.append((node, snk))
                plugins += snk.child("plugins").child("object").to_obj()
            if node.name in cls._tags :
                modname, clsname = cls._tags[node.name]
                if modname.startswith("snakes.plugins.") :
                    plugins.add(modname)
                elif node.name not in tag2obj :
                    tag2obj[node.name] = getattr(_snk_import(modname), clsname)
        for parent, child in trash :
            parent.children.remove(child)
        plugins.discard("snakes.nets")
        nets = snakes.plugins.load(plugins, "snakes.nets")
        for name, obj in inspect.getmembers(nets, inspect.isclass) :
            # Skip classes that cannot be serialised to PNML
            try :
                tag = obj.__pnmltag__
            except AttributeError :
                continue
            # Get the last class in the hierarchy that has the same
            # "__pnmltag__" and is in the same module. This is useful
            # for instance for snakes.typing.Type and its subclasses:
            # the former should be called used of the laters because
            # it dispatches the call to "__pnmlload__" according to
            # "__pnmltype__".
            bases = [obj] + [c for c in inspect.getmro(obj)
                             if (c.__module__ == obj.__module__)
                             and hasattr(c, "__pnmltag__")
                             and c.__pnmltag__ == tag]
            tag2obj[tag] = bases[-1]
        return result
    def nodes (self) :
        """Iterate over all the nodes (top-down) in a tree

        >>> t = Tree('foo', None,
        ...          Tree('bar', None),
        ...          Tree('egg', None,
        ...               Tree('spam', None)))
        >>> for node in t.nodes() :
        ...     print(str(node))
        <PNML tree 'foo'>
        <PNML tree 'bar'>
        <PNML tree 'egg'>
        <PNML tree 'spam'>

        @return: an iterator on all the nodes in the tree, including
            this one
        @rtype: `generator`
        """
        yield self
        for child in self.children :
            for node in child.nodes() :
                yield node
    def update (self, other) :
        """Incorporates children, attributes and data from another PNML
        tree

        >>> t = Tree('foo', 'hello',
        ...          Tree('bar', None),
        ...          Tree('egg', None,
        ...               Tree('spam', None)))
        >>> o = Tree('foo', 'world',
        ...          Tree('python', None),
        ...          attr='value')
        >>> t.update(o)
        >>> t
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <foo attr="value">
          <bar/>
          <egg>
           <spam/>
          </egg>
          <python/>
          hello
          world
         </foo>
        </pnml>
        >>> o = Tree('oops', None,
        ...          Tree('hello', None),
        ...          attr='value')
        >>> try : t.update(o)
        ... except SnakesError : print(sys.exc_info()[1])
        tag mismatch 'foo', 'oops'

        @param other: the other tree to get data from
        @type other: `Tree`
        @raise SnakesError: when `other` has not the same tag as
            `self`
        """
        if self.name != other.name :
            raise SnakesError("tag mismatch '%s', '%s'" % (self.name, other.name))
        self.children.extend(other.children)
        self.attributes.update(other.attributes)
        self.add_data(other.data)
    def add_child (self, child) :
        """Add a child to a PNML tree

        >>> t = Tree('foo', None)
        >>> t.add_child(Tree('bar', None))
        >>> t
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <foo>
          <bar/>
         </foo>
        </pnml>

        @param child: the PNML tree to append
        @type child: `Tree`
        """
        self.children.append(child)
    def add_data (self, data, sep='\n') :
        """Appends data to the current node

        >>> t = Tree('foo', None)
        >>> t.add_data('hello')
        >>> t
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <foo>
          hello
         </foo>
        </pnml>
        >>> t.add_data('world')
        >>> t
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <foo>
          hello
          world
         </foo>
        </pnml>
        >>> t.add_data('!', '')
        >>> t
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <foo>
          hello
          world!
         </foo>
        </pnml>

        @param data: the data to add
        @type data: `str`
        @param sep: separator to insert between pieces of data
        @type sep: `str`
        """
        try :
            data = data.strip()
            if data != "" :
                if self.data is None :
                    self.data = data
                else :
                    self.data += sep + data
        except :
            pass
    def __getitem__ (self, name) :
        """Returns one attribute

        >>> Tree('foo', None, x='egg', y='spam')['x']
        'egg'
        >>> Tree('foo', None, x='egg', y='spam')['z']
        Traceback (most recent call last):
          ...
        KeyError: 'z'

        @param name: the name of the attribute
        @type name: `str`
        @return: the value of the attribute
        @rtype: `str`
        @raise KeyError: if no such attribute is found
        """
        return self.attributes[name]
    def __setitem__ (self, name, value) :
        """Sets an attribute

        >>> t = Tree('foo', None)
        >>> t['egg'] = 'spam'
        >>> t
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <foo egg="spam"/>
        </pnml>

        @param name: the name of the attribute
        @type name: `str`
        @param value: the value of the attribute
        @type value: `str`
        """
        self.attributes[name] = value
    def __iter__ (self) :
        """Iterate over children nodes

        >>> [str(node) for node in Tree('foo', None,
        ...                             Tree('egg', None),
        ...                             Tree('spam', None,
        ...                                  Tree('bar', None)))]
        ["<PNML tree 'egg'>", "<PNML tree 'spam'>"]

        @return: an iterator over direct children of the node
        @rtype: `generator`
        """
        return iter(self.children)
    def has_child (self, name) :
        """Test if the tree has the given tag as a direct child

        >>> t = Tree('foo', None,
        ...          Tree('egg', None),
        ...          Tree('spam', None,
        ...               Tree('bar', None)))
        >>> t.has_child('egg')
        True
        >>> t.has_child('bar')
        False
        >>> t.has_child('python')
        False

        @param name: tag name to search for
        @type name: `str`
        @return: a Boolean value indicating wether such a child was
            found or not
        @rtype: `bool`
        """
        for child in self :
            if child.name == name :
                return True
        return False
    def child (self, name=None) :
        """Return the direct child that as the given tag

        >>> t = Tree('foo', None,
        ...          Tree('egg', 'first'),
        ...          Tree('egg', 'second'),
        ...          Tree('spam', None,
        ...               Tree('bar', None)))
        >>> t.child('spam')
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <spam>
          <bar/>
         </spam>
        </pnml>
        >>> try : t.child('python')
        ... except SnakesError : print(sys.exc_info()[1])
        no child 'python'
        >>> try : t.child('bar')
        ... except SnakesError : print(sys.exc_info()[1])
        no child 'bar'
        >>> try : t.child('egg')
        ... except SnakesError : print(sys.exc_info()[1])
        multiple children 'egg'
        >>> try : t.child()
        ... except SnakesError : print(sys.exc_info()[1])
        multiple children

        @param name: name of the tag to search for, if `None`, the
            fisrt child is returned if it is the only child
        @type name: `str` or `None`
        @return: the only child with the given name, or the only child
            if no name is given
        @rtype: `Tree`
        @raise SnakesError: when no child or more than one child could
            be returned
        """
        result = None
        for child in self :
            if name is None or child.name == name :
                if result is None :
                    result = child
                elif name is None :
                    raise SnakesError("multiple children")
                else :
                    raise SnakesError("multiple children '%s'" % name)
        if result is None :
            raise SnakesError("no child '%s'" % name)
        return result
    def get_children (self, name=None) :
        """Iterates over direct children having the given tag

        >>> t = Tree('foo', None,
        ...          Tree('egg', 'first'),
        ...          Tree('egg', 'second'),
        ...          Tree('spam', None,
        ...               Tree('bar', None)))
        >>> [str(n) for n in t.get_children()]
        ["<PNML tree 'egg'>", "<PNML tree 'egg'>", "<PNML tree 'spam'>"]
        >>> [str(n) for n in t.get_children('egg')]
        ["<PNML tree 'egg'>", "<PNML tree 'egg'>"]
        >>> [str(n) for n in t.get_children('python')]
        []
        >>> [str(n) for n in t.get_children('bar')]
        []

        @param name: tag to search for or `None`
        @type name: `str` or `None`
        @return: iterator over all the children if `name` is `None`,
            or over the children with tag `name` otherwise
        @rtype: `generator`
        """
        for child in self :
            if name is None or child.name == name :
                yield child
    def __str__ (self) :
        """Return a simple string representation of the node

        >>> str(Tree('foo', None, Tree('child', None)))
        "<PNML tree 'foo'>"

        @return: simple string representation of the node
        @rtype: `str`
        """
        return "<PNML tree %r>" % self.name
    def __repr__ (self) :
        """Return a detailed representation of the node.

        This is actually the XML text that corresponds to the `Tree`,
        as returned by `Tree.to_pnml`.

        >>> print(repr(Tree('foo', None, Tree('child', None))))
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <foo>
          <child/>
         </foo>
        </pnml>

        @return: XML string representation of the node
        @rtype: `str`
        """
        return self.to_pnml()
    _elementary = set(("str", "int", "float", "bool"))
    _collection = set(("list", "tuple", "set"))
    @classmethod
    def from_obj (cls, obj) :
        """Builds a PNML tree from an object.

        Objects defined in SNAKES usually have a method `__pnmldump__`
        that handles the conversion, for instance:

        >>> import snakes.nets
        >>> Tree.from_obj(snakes.nets.Place('p'))
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <place id="p">
          <type domain="universal"/>
          <initialMarking>
           <multiset/>
          </initialMarking>
         </place>
        </pnml>

        Most basic Python classes are handled has readable XML:

        >>> Tree.from_obj(42)
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <object type="int">
          42
         </object>
        </pnml>
        >>> Tree.from_obj([1, 2, 3])
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <object type="list">
          <object type="int">
           1
          </object>
          <object type="int">
           2
          </object>
          <object type="int">
           3
          </object>
         </object>
        </pnml>

        Otherwise, the object is serialised using module `pickle`,
        which allows to embed almost anything into PNML.

        >>> import re
        >>> Tree.from_obj(re.compile('foo|bar')) # serialized data replaced with '...'
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <object type="pickle">
          ...
         </object>
        </pnml>

        @param obj: the object to convert to PNML
        @type obj: `object`
        @return: the corresponding PNML tree
        @rtype: `Tree`
        """
        try :
            result = obj.__pnmldump__()
            result._tag2obj = {result.name : obj}
            if hasattr(obj, "__plugins__") :
                result._plugins = obj.__plugins__
            return result
        except AttributeError :
            pass
        result = Tree("object", None)
        _type = type(obj)
        _name = _type.__name__
        if _name in cls._elementary :
            handler = result._set_elementary
        elif _name in cls._collection :
            handler = result._set_collection
        elif inspect.ismethod(obj) :
            handler = result._set_method
            _name = "method"
        elif inspect.isclass(obj) :
            handler = result._set_class
            _name = "class"
        elif inspect.isroutine(obj) :
            handler = result._set_function
            _name = "function"
        elif inspect.ismodule(obj) :
            handler = result._set_module
            _name = "module"
        else :
            try :
                handler = getattr(result, "_set_" + _name)
            except AttributeError :
                handler = result._set_pickle
                _name = "pickle"
        result["type"] = _name
        handler(obj)
        return result
    def _set_NoneType (self, value) :
        pass
    def _get_NoneType (self) :
        pass
    def _set_elementary (self, value) :
        self.data = str(value)
    def _get_elementary (self) :
        if self["type"] == "bool" :
            return self.data.strip() == "True"
        return getattr(builtins, self["type"])(self.data)
    def _set_collection (self, value) :
        for v in value :
            self.add_child(self.from_obj(v))
    def _get_collection (self) :
        return getattr(builtins, self["type"])(child.to_obj()
                                               for child in self)
    def _set_dict (self, value) :
        for k, v in value.items() :
            self.add_child(Tree("item", None,
                                Tree("key", None, self.from_obj(k)),
                                Tree("value", None, self.from_obj(v))))
    def _get_dict (self) :
        return dict((child.child("key").child("object").to_obj(),
                     child.child("value").child("object").to_obj())
                    for child in self)
    def _native (self, obj) :
        try :
            if (obj.__module__ in ("__builtin__", "__builtins__", "builtins")
                or obj.__module__ == "snakes"
                or obj.__module__.startswith("snakes")
                or inspect.isbuiltin(obj)) :
                return True
        except :
            pass
        try :
            lib = os.path.dirname(inspect.getfile(inspect.getfile))
            if os.path.isfile(lib) :
                lib = os.path.dirname(lib)
            lib += os.sep
            try :
                path = inspect.getfile(obj)
            except :
                path = inspect.getmodule(obj).__file__
            return path.startswith(lib)
        except :
            return False
    def _name (self, obj) :
        try :
            name = obj.__module__
        except :
            name = inspect.getmodule(obj).__name__
        if name in ("__builtin__", "__builtins__", "builtins") :
            return obj.__name__
        else :
            return name + "." + obj.__name__
    def _set_class (self, value) :
        if self._native(value) :
            self["name"] = self._name(value)
        else :
            self._set_pickle(value)
    def _get_class (self) :
        if self.data :
            return self._get_pickle()
        elif "." in self["name"] :
            module, name = self["name"].rsplit(".", 1)
            return getattr(__import__(module, fromlist=[name]), name)
        else :
            return getattr(builtins, self["name"])
    def _set_function (self, value) :
        self._set_class(value)
    def _get_function (self) :
        return self._get_class()
    def _set_method (self, value) :
        self._set_function(value)
        self["name"] = "%s.%s" % (self["name"], value.__name__)
    def _get_method (self) :
        if self.data :
            return self._get_pickle()
        module, cls, name = self["name"].rsplit(".", 2)
        cls = getattr(__import__(module, fromlist=[name]), name)
        return getattr(cls, name)
    def _set_module (self, value) :
        self["name"] = value.__name__
    def _get_module (self) :
        return __import__(self["name"], fromlist=["__name__"])
    def _set_pickle (self, value) :
        self["type"] = "pickle"
        self.data = pickle.dumps(value)
        if PY3 :
            self.data = repr(self.data)
    def _get_pickle (self) :
        if PY3 :
            return pickle.loads(ast.literal_eval(self.data))
        else :
            return pickle.loads(self.data)
    def to_obj (self) :
        """Build an object from its PNML representation

        This is just the reverse as `Tree.from_obj`, objects that have
        a `__pnmldump__` method should also have a `__pnmlload__`
        class method to perform the reverse operation, together with
        an attribute `__pnmltag__`. Indeed, when a PNML node with tag
        name 'foo' has to be converted to an object, a class `C` such
        that `C.__pnmltag__ == 'foo'` is searched in module
        `snakes.nets` and `C.__pnmlload__(tree)` is called to rebuild
        the object.

        Standard Python objects and pickled ones are also recognised
        and correctly rebuilt.

        >>> import snakes.nets
        >>> Tree.from_obj(snakes.nets.Place('p')).to_obj()
        Place('p', MultiSet([]), tAll)
        >>> Tree.from_obj(42).to_obj()
        42
        >>> Tree.from_obj([1, 2, 3]).to_obj()
        [1, 2, 3]
        >>> import re
        >>> Tree.from_obj(re.compile('foo|bar')).to_obj()
        <... object at ...>

        @return: the Python object encoded by the PNML tree
        @rtype: `object`
        """
        if self.name == "pnml" :
            if len(self.children) == 0 :
                raise SnakesError("empty PNML content")
            elif len(self.children) == 1 :
                return self.child().to_obj()
            else :
                return tuple(child.to_obj() for child in self.children
                             if child.name != "snakes")
        elif self.name == "object" :
            if self["type"] in self._elementary :
                handler = self._get_elementary
            elif self["type"] in self._collection :
                handler = self._get_collection
            else :
                try :
                    handler = getattr(self, "_get_" + self["type"])
                except AttributeError :
                    handler = self._get_pickle
            return handler()
        elif self.name != "snakes" :
            try :
                if self.name in self._tag2obj :
                    return self._tag2obj[self.name].__pnmlload__(self)
            except AttributeError :
                pass
        raise SnakesError("unsupported PNML tag '%s'" % self.name)

def dumps (obj) :
    """Dump an object to a PNML string, any Python object may be
    serialised this way (resorting to pickling when the object does
    not support serialisation to PNML).

    >>> print(dumps(42))
    <?xml version="1.0" encoding="utf-8"?>
    <pnml>
     <object type="int">
      42
     </object>
    </pnml>

    @param obj: the object to dump
    @type obj: `object`
    @return: the PNML that represents the object
    @rtype: `str`
    """
    return Tree.from_obj(obj).to_pnml()

def loads (source, plugins=[]) :
    """Load an object from a PNML string

    >>> loads(dumps(42))
    42

    @param source: the data to parse
    @type source: `str`
    @return: the object represented by the source
    @rtype: `object`
    """
    return Tree.from_pnml(source, plugins).to_obj()



import sys, os.path, cgi, functools, mimetypes
import os, traceback, random, base64, inspect
from snakes.utils.simul import logger as log
from snakes.utils.simul.html import json, utf8

try :
    import httplib
except :
    import http as httplib

try :
    import urlparse
except :
    import urllib as urlparse

try :
    import BaseHTTPServer
except :
    import http as BaseHTTPServer

##
##
##

class HTTPError (Exception) :
    def __init__ (self, code, reason=None, debug=None, headers={}) :
        self.answer = httplib.responses[code]
        if reason is None :
            message = self.answer
        else :
            message = "%s (%s)" % (self.answer, reason)
        Exception.__init__(self, message)
        self.code = code
        self.reason = reason
        self.debug = debug
        self.headers = headers

##
##
##

encoders = {"application/json" : json,
            "application/binary" : str,
            "text/plain" : utf8,
            "text/html" : utf8,
            }

def http (content_type=None, **types) :
    def decorator (method) :
        @functools.wraps(method)
        def wrapper (*larg, **karg) :
            try :
                args = inspect.getcallargs(method, *larg, **karg)
                for a, t in types.items() :
                    if a in args :
                        args[a] = t(args[a])
            except :
                raise HTTPError(httplib.BAD_REQUEST, "invalid arguments")
            try :
                if content_type is None :
                    return method(**args)
                else :
                    encode = encoders.get(content_type, str)
                    return content_type, encode(method(**args))
            except HTTPError :
                raise
            except :
                raise HTTPError(httplib.INTERNAL_SERVER_ERROR,
                                debug=sys.exc_info())
        return wrapper
    return decorator

class Node (object) :
    def __init__ (self, **children) :
        for child, node in children.items() :
            setattr(self, child, node)
    def __getitem__ (self, path) :
        path = path.strip("/")
        if "/" in path :
            head, tail = path.split("/", 1)
            child = getattr(self, head, None)
            if isinstance(child, Node) :
                return child[tail]
            else :
                raise KeyError(tail)
        elif path == "" and hasattr(self, "__call__") :
            return self.__call__
        elif hasattr(self, path) :
            return getattr(self, path)
        else :
            raise KeyError(path)

class DirNode (Node) :
    def __init__ (self, path) :
        self.root = os.path.realpath(path)
    def __getitem__ (self, path) :
        path = os.path.join(self.root, path.lstrip("./"))
        if not os.path.isfile(path) :
            raise HTTPError(httplib.NOT_FOUND)
        ct = mimetypes.guess_type(path)[0] or "application/octet-stream"
        @http(ct)
        def handler () :
            return open(path).read()
        return handler

class ResourceNode (Node) :
    def __init__ (self, data, dirs) :
        self.data = data
        self.dirs = dirs
        self.ct = dict((path, mimetypes.guess_type(path)[0]
                        or "application/octet-stream")
                       for path in self.data)
    def __getitem__ (self, path) :
        if path in self.data :
            @http(self.ct[path])
            def handler () :
                return self.data[path]
            return handler
        else :
            try :
                base, child = path.split("/", 1)
                return self.dirs[base][child]
            except :
                raise HTTPError(httplib.NOT_FOUND)


##
##
##

class HTTPRequestHandler (BaseHTTPServer.BaseHTTPRequestHandler) :
    def do_GET (self) :
        url = self.geturl()
        self.do(url, url.query)
    def do_POST (self) :
        content_len = int(self.headers.getheader('content-length'))
        data = self.rfile.read(content_len)
        self.do(self.geturl(), data)
    def geturl (self) :
        try :
            return urlparse.urlparse(self.path)
        except :
            raise HTTPError(httplib.BAD_REQUEST, "invalid URL")
    def do (self, url, data) :
        try :
            # warning: parse_qs returns lists for values => use parse_qsl
            query = dict(cgi.parse_qsl(data))
            # jQuery may add _ in query for cache control, let's drop it
            query.pop("_", None)
        except :
            raise HTTPError(httplib.BAD_REQUEST, "invalid query")
        try :
            try :
                handler = self.server[url.path]
            except KeyError :
                raise HTTPError(httplib.NOT_FOUND)
            content_type, data = handler(**query)
            self.send_response(httplib.OK)
            self.send_header("Content-type", content_type)
            self.send_header("Content-length", len(data))
            self.end_headers()
            self.wfile.write(data)
        except HTTPError :
            c, v, t = sys.exc_info()
            self.send_response(v.code)
            self.send_header("Content-type", "text/html")
            for hname, hdata in v.headers.iteritems() :
                self.send_header(hname, hdata)
            self.end_headers()
            self.wfile.write("<html><title>%s</title></head>"
                             "<body><p>%s</p></body>" % (v.answer, v.message))
            if v.code == 500 :
                traceback.print_exception(*v.debug)
    def log_request (self, code="-", size="-") :
        code = str(code or "-")
        method, path, version = self.requestline.split()
        if code[0] in "45" :
            logger = log.warn
        else :
            path = path.split("/", 2)[-1]
            if len(path) > 28 :
                path = "..." + path[-28:]
            logger = log.debug
        logger("%s %s => %s" % (method, path, code), "httpd")
    def log_message (self, format, *args) :
        log.info(format % args, "httpd")
    def log_error (self, format, *args) :
        log.error(format % args, "httpd")

class HTTPServer (BaseHTTPServer.HTTPServer):
    def __init__ (self, server_address, root):
        BaseHTTPServer.HTTPServer.__init__(self, server_address,
                                           HTTPRequestHandler)
        self.root = root
        self.key = "".join(base64.b64encode(
            "".join(chr(random.getrandbits(8)) for i in range(15)),
            "-_").split())
    def __getitem__ (self, path) :
        try :
            key, path = path.lstrip("/").split("/", 1)
        except :
            raise HTTPError(httplib.FORBIDDEN)
        if key != self.key :
            raise HTTPError(httplib.FORBIDDEN)
        return self.root[path]

##
##
##

if __name__ == '__main__':
    class HelloNode (Node) :
        @http("text/plain")
        def hello (self, first, last) :
            return "Hello %s %s!" % (first.capitalize(), last.capitalize())
    try :
        httpd = HTTPServer(('', 1234), HelloNode(r=DirNode(".")))
        httpd.serve_forever()
    except KeyboardInterrupt :
        print("\rGoobye")



"""
>>> testparser(Translator)
"""

import operator, sys
import snakes
from snakes.lang.python.parser import (ParseTree, ParseTestParser,
                                       Translator as PyTranslator,
                                       ParseTree as PyParseTree,
                                       testparser)
from snakes.lang.pgen import ParseError
from snakes.lang.abcd.pgen import parser
import snakes.lang.abcd.asdl as ast

_symbols = parser.tokenizer.tok_name.copy()
# next statement overrides 'NT_OFFSET' entry with 'single_input'
# (this is desired)
_symbols.update(parser.symbolMap)

def skip (token) :
    if token.kind == token.lexer.COMMENT :
        words = token.strip().split()
        if words[:2] == ["#", "coding="] :
            coding = words[2]
        elif words[:3] == ["#", "-*-", "coding:"] :
            coding = words[3]
        else :
            return
        snakes.defaultencoding = coding

parser.tokenizer.skip_token = skip

class ParseTree (PyParseTree) :
    _symbols = _symbols

class Translator (PyTranslator) :
    ParseTree = ParseTree
    parser = parser
    ST = ast
    def do_file_input (self, st, ctx) :
        """file_input: abcd_main ENDMARKER
        -> ast.AbcdSpec

        <<< symbol FOO, BAR
        ... [egg+(spam) if spam == ham] ; [spam-(FOO, BAR)]
        "AbcdSpec(context=[AbcdSymbol(symbols=['FOO', 'BAR'])], body=AbcdFlowOp(left=AbcdAction(accesses=[SimpleAccess(buffer='egg', arc=Produce(), tokens=Name(id='spam', ctx=Load()))], guard=Compare(left=Name(id='spam', ctx=Load()), ops=[Eq()], comparators=[Name(id='ham', ctx=Load())])), op=Sequence(), right=AbcdAction(accesses=[SimpleAccess(buffer='spam', arc=Consume(), tokens=Tuple(elts=[Name(id='FOO', ctx=Load()), Name(id='BAR', ctx=Load())], ctx=Load()))], guard=Name(id='True', ctx=Load()))), asserts=[])"
        """
        return self.do(st[0])
    def do_abcd_main (self, st, ctx) :
        """abcd_main: (NEWLINE | abcd_global)* abcd_expr (abcd_prop)*
        -> ast.AbcdSpec

        <<< symbol FOO, BAR
        ... [egg+(spam) if spam == ham] ; [spam-(FOO, BAR)]
        "AbcdSpec(context=[AbcdSymbol(symbols=['FOO', 'BAR'])], body=AbcdFlowOp(left=AbcdAction(accesses=[SimpleAccess(buffer='egg', arc=Produce(), tokens=Name(id='spam', ctx=Load()))], guard=Compare(left=Name(id='spam', ctx=Load()), ops=[Eq()], comparators=[Name(id='ham', ctx=Load())])), op=Sequence(), right=AbcdAction(accesses=[SimpleAccess(buffer='spam', arc=Consume(), tokens=Tuple(elts=[Name(id='FOO', ctx=Load()), Name(id='BAR', ctx=Load())], ctx=Load()))], guard=Name(id='True', ctx=Load()))), asserts=[])"
        """
        expr = len(st)-1
        for i, child in enumerate(st) :
            if child.symbol == "abcd_expr" :
                expr = i
                break
        return self.ST.AbcdSpec(lineno=st.srow, col_offset=st.scol,
                                context=[self.do(child) for child in st[:expr]
                                         if child.kind != self.NEWLINE],
                                body=self.do(st[expr]),
                                asserts=[self.do(child) for child in st[expr+1:]])
    def do_abcd_prop (self, st, ctx) :
        """abcd_prop: 'assert' test (NEWLINE)*
        -> ast.test

        <<< [True]
        ... assert True
        ... assert False
        "AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[Name(id='True', ctx=Load()), Name(id='False', ctx=Load())])"
        """
        return self.do(st[1])
    def do_abcd_global (self, st, ctx) :
        """abcd_global: (import_stmt | abcd_symbol | abcd_typedef
               | abcd_const | abcd_decl)
        -> ast.AST

        <<< import module
        ... from module import content
        ... symbol EGG, SPAM, HAM
        ... typedef t : enum(EGG, SPAM, HAM)
        ... net Foo() : [True]
        ... buffer bar : t = ()
        ... [True]
        "AbcdSpec(context=[Import(names=[alias(name='module', asname=None)]), ImportFrom(module='module', names=[alias(name='content', asname=None)], level=0), AbcdSymbol(symbols=['EGG', 'SPAM', 'HAM']), AbcdTypedef(name='t', type=EnumType(items=[Name(id='EGG', ctx=Load()), Name(id='SPAM', ctx=Load()), Name(id='HAM', ctx=Load())])), AbcdNet(name='Foo', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])), AbcdBuffer(name='bar', type=NamedType(name='t'), capacity=None, content=Tuple(elts=[], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        return self.do(st[0])
    def do_abcd_spec (self, st, ctx) :
        """abcd_spec: (NEWLINE | abcd_decl)* abcd_expr
        -> ast.AbcdSpec

        <<< net Foo () :
        ...     buffer bar : spam = ()
        ...     net Bar () : [False]
        ...     [True]
        ... Foo()
        "AbcdSpec(context=[AbcdNet(name='Foo', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[AbcdBuffer(name='bar', type=NamedType(name='spam'), capacity=None, content=Tuple(elts=[], ctx=Load())), AbcdNet(name='Bar', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=False), asserts=[]))], body=AbcdAction(accesses=[], guard=True), asserts=[]))], body=AbcdInstance(net='Foo', asname=None, args=[], keywords=[], starargs=None, kwargs=None), asserts=[])"
        """
        tree = self.do_abcd_main(st, ctx)
        tree.st = st
        return tree
    def do_abcd_decl (self, st, ctx) :
        """abcd_decl: abcd_net | abcd_task | abcd_buffer
        -> ast.AST

        <<< net Foo () :
        ...     buffer bar : spam = ()
        ...     net Bar () : [False]
        ...     [True]
        ... Foo()
        "AbcdSpec(context=[AbcdNet(name='Foo', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[AbcdBuffer(name='bar', type=NamedType(name='spam'), capacity=None, content=Tuple(elts=[], ctx=Load())), AbcdNet(name='Bar', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=False), asserts=[]))], body=AbcdAction(accesses=[], guard=True), asserts=[]))], body=AbcdInstance(net='Foo', asname=None, args=[], keywords=[], starargs=None, kwargs=None), asserts=[])"
        """
        tree = self.do_abcd_global(st, ctx)
        tree.st = st
        return tree
    def do_abcd_const (self, st, ctx) :
        """abcd_const: 'const' NAME '=' testlist
        -> ast.AbcdConst

        <<< const FOO = 5
        ... [True]
        "AbcdSpec(context=[AbcdConst(name='FOO', value=Num(n=5))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< const FOO = 1, 2, 3
        ... [True]
        "AbcdSpec(context=[AbcdConst(name='FOO', value=Tuple(elts=[Num(n=1), Num(n=2), Num(n=3)], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        return self.ST.AbcdConst(lineno=st.srow, col_offset=st.scol,
                                 name=st[1].text, value=self.do(st[3], ctx))
    def do_abcd_symbol (self, st, ctx) :
        """abcd_symbol: 'symbol' abcd_namelist
        -> ast.AbcdSymbol

        <<< symbol FOO
        ... [True]
        "AbcdSpec(context=[AbcdSymbol(symbols=['FOO'])], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< symbol FOO, BAR
        ... [True]
        "AbcdSpec(context=[AbcdSymbol(symbols=['FOO', 'BAR'])], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        return self.ST.AbcdSymbol(lineno=st.srow, col_offset=st.scol,
                                  symbols=self.do(st[1]))
    def do_abcd_namelist (self, st, ctx) :
        """abcd_namelist: NAME (',' NAME)*
        -> str+

        <<< symbol FOO
        ... [True]
        "AbcdSpec(context=[AbcdSymbol(symbols=['FOO'])], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< symbol FOO, BAR
        ... [True]
        "AbcdSpec(context=[AbcdSymbol(symbols=['FOO', 'BAR'])], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        return [child.text for child in st[::2]]
    def _do_flowop (self, st, op) :
        nodes = [self.do(child) for child in st[::2]]
        while len(nodes) > 1 :
            left = nodes.pop(0)
            right = nodes.pop(0)
            theop = op()
            theop.st = st[1]
            flow = self.ST.AbcdFlowOp(lineno=left.lineno,
                                      col_offset=left.col_offset,
                                      left=left, op=theop, right=right)
            flow.st = st
            nodes.insert(0, flow)
        return nodes[0]
    def do_abcd_expr (self, st, ctx) :
        """abcd_expr: abcd_choice_expr ('|' abcd_choice_expr)*
        -> ast.process

        <<< [True]
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])'
        <<< [True] | [False]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Parallel(), right=AbcdAction(accesses=[], guard=False)), asserts=[])'
        <<< [True] | [False] | [True]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Parallel(), right=AbcdAction(accesses=[], guard=False)), op=Parallel(), right=AbcdAction(accesses=[], guard=True)), asserts=[])'

        """
        return self._do_flowop(st, self.ST.Parallel)
    def do_abcd_choice_expr (self, st, ctx) :
        """abcd_choice_expr: abcd_iter_expr ('+' abcd_iter_expr)*
        -> ast.process

        <<< [True]
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])'
        <<< [True] + [False]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Choice(), right=AbcdAction(accesses=[], guard=False)), asserts=[])'
        <<< [True] + [False] + [True]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Choice(), right=AbcdAction(accesses=[], guard=False)), op=Choice(), right=AbcdAction(accesses=[], guard=True)), asserts=[])'
        """
        return self._do_flowop(st, self.ST.Choice)
    def do_abcd_iter_expr (self, st, ctx) :
        """abcd_iter_expr: abcd_seq_expr ('*' abcd_seq_expr)*
        -> ast.process

        <<< [True]
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])'
        <<< [True] * [False]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Loop(), right=AbcdAction(accesses=[], guard=False)), asserts=[])'
        <<< [True] * [False] * [True]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Loop(), right=AbcdAction(accesses=[], guard=False)), op=Loop(), right=AbcdAction(accesses=[], guard=True)), asserts=[])'
        """
        return self._do_flowop(st, self.ST.Loop)
    def do_abcd_seq_expr (self, st, ctx) :
        """abcd_seq_expr: abcd_base_expr (';' abcd_base_expr)*
        -> ast.process

        <<< [True]
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])'
        <<< [True] ; [False]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Sequence(), right=AbcdAction(accesses=[], guard=False)), asserts=[])'
        <<< [True] ; [False] ; [True]
        'AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Sequence(), right=AbcdAction(accesses=[], guard=False)), op=Sequence(), right=AbcdAction(accesses=[], guard=True)), asserts=[])'
        """
        return self._do_flowop(st, self.ST.Sequence)
    def do_abcd_base_expr (self, st, ctx) :
        """abcd_base_expr: (abcd_action | '(' abcd_expr ')') (NEWLINE)*
        -> ast.process

        <<< [True]
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])'
        <<< ([True])
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])'
        """
        if st[0].text == "(" :
            return self.do(st[1])
        else :
            return self.do(st[0])
    def do_abcd_action (self, st, ctx) :
        """abcd_action: ('[' 'True' ']' | '[' 'False' ']' |
                         '[' abcd_access_list ['if' test] ']' |
                         abcd_instance)
        -> ast.AbcdAction | ast.AbcdInstance

        <<< [True]
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[])'
        <<< [False]
        'AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=False), asserts=[])'
        <<< Foo(1, 2)
        "AbcdSpec(context=[], body=AbcdInstance(net='Foo', asname=None, args=[Num(n=1), Num(n=2)], keywords=[], starargs=None, kwargs=None), asserts=[])"
        """
        if len(st) == 1 :
            return self.do(st[0])
        elif st[1].text == "True" :
            return self.ST.AbcdAction(lineno=st.srow, col_offset=st.scol,
                                      accesses=[], guard=True)
        elif st[1].text == "False" :
            return self.ST.AbcdAction(lineno=st.srow, col_offset=st.scol,
                                      accesses=[], guard=False)
        elif len(st) == 3 :
            return self.ST.AbcdAction(lineno=st.srow, col_offset=st.scol,
                                      accesses=self.do(st[1]),
                                      guard=self.ST.Name(lineno=st[-1].srow,
                                                         col_offset=st[-1].scol,
                                                         id="True",
                                                         ctx=self.ST.Load()))
        else :
            return self.ST.AbcdAction(lineno=st.srow, col_offset=st.scol,
                                      accesses=self.do(st[1]),
                                      guard=self.do(st[3]))
    def do_abcd_access_list (self, st, ctx) :
        """abcd_access_list: abcd_access (',' abcd_access)*
        -> ast.access+

        <<< [foo-(1)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[SimpleAccess(buffer='foo', arc=Consume(), tokens=Num(n=1))], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [foo-(1), bar+(2)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[SimpleAccess(buffer='foo', arc=Consume(), tokens=Num(n=1)), SimpleAccess(buffer='bar', arc=Produce(), tokens=Num(n=2))], guard=Name(id='True', ctx=Load())), asserts=[])"
        """
        return [self.do(child) for child in st[::2]]
    _arc = {"+" : ast.Produce,
            "-" : ast.Consume,
            "?" : ast.Test,
            "<<" : ast.Fill}
    def do_abcd_access (self, st, ctx) :
        """abcd_access: (NAME '+' '(' testlist ')' |
              NAME '?' '(' testlist ')' |
              NAME '-' '(' testlist ')' |
              NAME '<>' '(' testlist '=' testlist ')' |
              NAME '>>' '(' NAME ')' |
              NAME '<<' '(' testlist_comp ')' |
              NAME '.' NAME '(' test (',' test)* ')')
        -> ast.access

        <<< [egg+(x), spam-(y), ham?(z), foo<<(bar)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[SimpleAccess(buffer='egg', arc=Produce(), tokens=Name(id='x', ctx=Load())), SimpleAccess(buffer='spam', arc=Consume(), tokens=Name(id='y', ctx=Load())), SimpleAccess(buffer='ham', arc=Test(), tokens=Name(id='z', ctx=Load())), SimpleAccess(buffer='foo', arc=Fill(), tokens=Name(id='bar', ctx=Load()))], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [foo<<(spam(egg) for egg in ham)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[SimpleAccess(buffer='foo', arc=Fill(), tokens=ListComp(elt=Call(func=Name(id='spam', ctx=Load()), args=[Name(id='egg', ctx=Load())], keywords=[], starargs=None, kwargs=None), generators=[comprehension(target=Name(id='egg', ctx=Store()), iter=Name(id='ham', ctx=Load()), ifs=[])]))], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [bar-(l), foo<<(l)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[SimpleAccess(buffer='bar', arc=Consume(), tokens=Name(id='l', ctx=Load())), SimpleAccess(buffer='foo', arc=Fill(), tokens=Name(id='l', ctx=Load()))], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [bar-(l), foo<<(l,)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[SimpleAccess(buffer='bar', arc=Consume(), tokens=Name(id='l', ctx=Load())), SimpleAccess(buffer='foo', arc=Fill(), tokens=Tuple(elts=[Name(id='l', ctx=Load())], ctx=Load()))], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [spam>>(ham) if spam is egg]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[FlushAccess(buffer='spam', target='ham')], guard=Compare(left=Name(id='spam', ctx=Load()), ops=[Is()], comparators=[Name(id='egg', ctx=Load())])), asserts=[])"
        <<< [count<>(n = n+1)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[SwapAccess(buffer='count', target=Name(id='n', ctx=Load()), tokens=BinOp(left=Name(id='n', ctx=Load()), op=Add(), right=Num(n=1)))], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [foo.spawn(child, 1, 2, 3)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[Spawn(net='foo', pid=Name(id='child', ctx=Load()), args=[Num(n=1), Num(n=2), Num(n=3)])], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [foo.wait(child, y, z)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[Wait(net='foo', pid=Name(id='child', ctx=Load()), args=[Name(id='y', ctx=Load()), Name(id='z', ctx=Load())])], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [foo.suspend(pid)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[Suspend(net='foo', pid=Name(id='pid', ctx=Load()))], guard=Name(id='True', ctx=Load())), asserts=[])"
        <<< [foo.resume(pid)]
        "AbcdSpec(context=[], body=AbcdAction(accesses=[Resume(net='foo', pid=Name(id='pid', ctx=Load()))], guard=Name(id='True', ctx=Load())), asserts=[])"
        """
        if st[1].text in ("+", "?", "-") :
            return self.ST.SimpleAccess(lineno=st.srow, col_offset=st.scol,
                                        buffer=st[0].text,
                                        arc=self._arc[st[1].text](),
                                        tokens=self.do(st[3]))
        elif st[1].text == "<<" :
            loop, elts, atom = self.do(st[3], ctx)
            if atom is not None :
                args = atom
            elif loop is None :
                args = self.ST.Tuple(lineno=st.srow, col_offset=st.scol,
                                     elts=elts, ctx=ctx())
            else :
                args = self.ST.ListComp(lineno=st.srow, col_offset=st.scol,
                                        elt=loop, generators=elts)
            return self.ST.SimpleAccess(lineno=st.srow, col_offset=st.scol,
                                        buffer=st[0].text,
                                        arc=self._arc[st[1].text](),
                                        tokens=args)
        elif st[1].text == ">>" :
            return self.ST.FlushAccess(lineno=st.srow, col_offset=st.scol,
                                       buffer=st[0].text,
                                       target=st[3].text)
        elif st[1].text == "<>" :
            return self.ST.SwapAccess(lineno=st.srow, col_offset=st.scol,
                                      buffer=st[0].text,
                                      target=self.do(st[3]),
                                      tokens=self.do(st[5]))
        elif st[2].text in ("suspend", "resume") : # st[1].text == "."
            if len(st) > 6 :
                raise ParseError(st.text, reason="too many arguments for %s"
                                 % st[2].text)
            if st[2].text == "suspend" :
                tree = self.ST.Suspend
            else :
                tree = self.ST.Resume
            return tree(lineno=st.srow, col_offset=st.scol,
                        net=st[0].text,
                        pid=self.do(st[4]))
        elif st[2].text in ("spawn", "wait") : # st[1].text == "."
            if len(st) > 6 :
                args = [self.do(child) for child in st[6:-1:2]]
            else :
                args = []
            if st[2].text == "spawn" :
                tree = self.ST.Spawn
            else :
                tree = self.ST.Wait
            return tree(lineno=st.srow, col_offset=st.scol,
                        net=st[0].text,
                        pid=self.do(st[4]),
                        args=args)
        else :
            raise ParseError(st[2].text, reason=("expected 'spawn', 'wait', "
                             "'suspend' or 'resume', but found '%s'")
                             % st[2].text)
    def do_abcd_instance (self, st, ctx) :
        """abcd_instance: [NAME ':' ':'] NAME '(' [arglist] ')'
        -> ast.AbcdInstance

        <<< sub()
        "AbcdSpec(context=[], body=AbcdInstance(net='sub', asname=None, args=[], keywords=[], starargs=None, kwargs=None), asserts=[])"
        <<< sub(1, 2, 3)
        "AbcdSpec(context=[], body=AbcdInstance(net='sub', asname=None, args=[Num(n=1), Num(n=2), Num(n=3)], keywords=[], starargs=None, kwargs=None), asserts=[])"
        <<< sub(1, *l)
        "AbcdSpec(context=[], body=AbcdInstance(net='sub', asname=None, args=[Num(n=1)], keywords=[], starargs=Name(id='l', ctx=Load()), kwargs=None), asserts=[])"
        <<< sub(1, **d)
        "AbcdSpec(context=[], body=AbcdInstance(net='sub', asname=None, args=[Num(n=1)], keywords=[], starargs=None, kwargs=Name(id='d', ctx=Load())), asserts=[])"
        <<< sub(a=1, b=2)
        "AbcdSpec(context=[], body=AbcdInstance(net='sub', asname=None, args=[], keywords=[keyword(arg='a', value=Num(n=1)), keyword(arg='b', value=Num(n=2))], starargs=None, kwargs=None), asserts=[])"
        <<< foo::sub()
        "AbcdSpec(context=[], body=AbcdInstance(net='sub', asname='foo', args=[], keywords=[], starargs=None, kwargs=None), asserts=[])"
        <<< foo::sub(1, 2)
        "AbcdSpec(context=[], body=AbcdInstance(net='sub', asname='foo', args=[Num(n=1), Num(n=2)], keywords=[], starargs=None, kwargs=None), asserts=[])"
        """
        if len(st) in (3, 6) :
            args, keywords, starargs, kwargs = [], [], None, None
        else :
            args, keywords, starargs, kwargs = self.do(st[-2])
        if st[1].text == ':' :
            net = st[3].text
            asname = st[0].text
        else :
            net = st[0].text
            asname = None
        return self.ST.AbcdInstance(lineno=st.srow, col_offset=st.scol,
                                    net=net, asname=asname, args=args,
                                    keywords=keywords, starargs=starargs,
                                    kwargs=kwargs)
    def do_abcd_net (self, st, ctx) :
        """abcd_net: 'net' NAME parameters ':' abcd_suite
        -> ast.AbcdNet

        <<< net Foo () : [True]
        ... [False]
        "AbcdSpec(context=[AbcdNet(name='Foo', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[]))], body=AbcdAction(accesses=[], guard=False), asserts=[])"
        <<< net Foo (x, y) : [True]
        ... [False]
        "AbcdSpec(context=[AbcdNet(name='Foo', args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[]))], body=AbcdAction(accesses=[], guard=False), asserts=[])"
        """
        params = self.do(st[2])
        return self.ST.AbcdNet(lineno=st.srow, col_offset=st.scol,
                               name=st[1].text,
                               args=params,
                               body=self.do(st[4]))
    def do_abcd_task (self, st, ctx) :
        """abcd_task: 'task' NAME typelist '-' '>' typelist ':' abcd_suite
        -> ast.AbcdTask

        <<< task Foo (int) -> () : [True]
        ... [False]
        "AbcdSpec(context=[AbcdTask(name='Foo', body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[]), input=[NamedType(name='int')], output=[])], body=AbcdAction(accesses=[], guard=False), asserts=[])"
        """
        return self.ST.AbcdTask(lineno=st.srow, col_offset=st.scol,
                                name=st[1].text,
                                body=self.do(st[-1]),
                                input=self.do(st[2]),
                                output=self.do(st[5]))
    def do_typelist (self, st, ctx) :
        """typelist: '(' [abcd_type (',' abcd_type)*] ')'
        -> ast.abcdtype*

        <<< task Foo () -> (int, int, int|bool) : [True]
        ... [False]
        "AbcdSpec(context=[AbcdTask(name='Foo', body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[]), input=[], output=[NamedType(name='int'), NamedType(name='int'), UnionType(types=[NamedType(name='int'), NamedType(name='bool')])])], body=AbcdAction(accesses=[], guard=False), asserts=[])"
        """
        return [self.do(child) for child in st[1:-1:2]]
    def do_abcd_suite (self, st, ctx) :
        """abcd_suite: abcd_expr | NEWLINE INDENT abcd_spec DEDENT
        -> ast.AbcdSpec

        <<< net Foo () :
        ...    [True]
        ... [False]
        "AbcdSpec(context=[AbcdNet(name='Foo', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[]))], body=AbcdAction(accesses=[], guard=False), asserts=[])"
        <<< net Foo () : ([True]
        ...     + [False])
        ... [False]
        "AbcdSpec(context=[AbcdNet(name='Foo', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdFlowOp(left=AbcdAction(accesses=[], guard=True), op=Choice(), right=AbcdAction(accesses=[], guard=False)), asserts=[]))], body=AbcdAction(accesses=[], guard=False), asserts=[])"
        """
        if len(st) == 1 :
            return self.ST.AbcdSpec(lineno=st.srow, col_offset=st.scol,
                                    context=[],
                                    body=self.do(st[0]))
        else :
            return self.do(st[2])
    def do_abcd_buffer (self, st, ctx) :
        """[ decorators ] 'buffer' NAME ['[' test ']'] ':' abcd_type '=' testlist
        -> ast.AbcdBuffer

        <<< buffer foo : int = ()
        ... [True]
        "AbcdSpec(context=[AbcdBuffer(name='foo', type=NamedType(name='int'), capacity=None, content=Tuple(elts=[], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< buffer foo : int = 1, 2
        ... [True]
        "AbcdSpec(context=[AbcdBuffer(name='foo', type=NamedType(name='int'), capacity=None, content=Tuple(elts=[Num(n=1), Num(n=2)], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< buffer foo : int|bool = ()
        ... [True]
        "AbcdSpec(context=[AbcdBuffer(name='foo', type=UnionType(types=[NamedType(name='int'), NamedType(name='bool')]), capacity=None, content=Tuple(elts=[], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< @capacity(max=5)
        ... buffer foo : int = ()
        ... [True]
        "AbcdSpec(context=[AbcdBuffer(name='foo', type=NamedType(name='int'), capacity=[None, Num(n=5)], content=Tuple(elts=[], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< @capacity(min=2)
        ... buffer foo : int = ()
        ... [True]
        "AbcdSpec(context=[AbcdBuffer(name='foo', type=NamedType(name='int'), capacity=[Num(n=2), None], content=Tuple(elts=[], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< @capacity(min=2, max=5)
        ... buffer foo : int = ()
        ... [True]
        "AbcdSpec(context=[AbcdBuffer(name='foo', type=NamedType(name='int'), capacity=[Num(n=2), Num(n=5)], content=Tuple(elts=[], ctx=Load()))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        if len(st) == 6 : # no decorator, no array
            return self.ST.AbcdBuffer(lineno=st.srow, col_offset=st.scol,
                                      name=st[1].text,
                                      type=self.do(st[3]),
                                      capacity=None,
                                      content=self.do(st[-1]))
        elif len(st) == 7 : # decorator, no array
            deco = self.do_buffer_decorators(st[0], ctx)
            return self.ST.AbcdBuffer(lineno=st.srow, col_offset=st.scol,
                                      name=st[2].text,
                                      type=self.do(st[4]),
                                      capacity=deco["capacity"],
                                      content=self.do(st[-1]))
        else :
            raise ParseError(st.text,
                             reason="arrays not (yet) supported")
    def do_buffer_decorators (self, st, ctx) :
        deco = {}
        for child in st :
            tree = self.do(child)
            if isinstance(tree, self.ST.Call) and tree.func.id == "capacity" :
                if tree.args or tree.starargs or tree.kwargs :
                    raise ParseError(child, reason="invalid parameters")
                min, max = None, None
                for kw in tree.keywords :
                    if kw.arg == "min" :
                        min = kw.value
                    elif kw.arg == "max" :
                        max = kw.value
                    else :
                        raise ParseError(child,
                                         reason="invalid parameter %r" % kw.arg)
                if min or max :
                    deco["capacity"] = [min, max]
                else :
                    deco["capacity"] = None
                continue
            raise ParseError(child, reason="invalid buffer decorator")
        return deco
    def do_abcd_typedef (self, st, ctx) :
        """abcd_typedef: 'typedef' NAME ':' abcd_type
        -> ast.AbcdTypedef

        <<< typedef foo : int
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=NamedType(name='int'))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        return self.ST.AbcdTypedef(lineno=st.srow, col_offset=st.scol,
                                   name=st[1].text,
                                   type=self.do(st[3]))
    def do_abcd_type (self, st, ctx) :
        """abcd_type: abcd_and_type ('|' abcd_and_type)*
        -> snakes.typing.Type

        <<< typedef foo : int
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=NamedType(name='int'))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< typedef foo : int | bool
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=UnionType(types=[NamedType(name='int'), NamedType(name='bool')]))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        if len(st) == 1 :
            return self.do(st[0])
        else :
            return self.ST.UnionType(lineno=st.srow, col_offset=st.scol,
                                     types=[self.do(child) for child in st[::2]])
    def do_abcd_and_type (self, st, ctx) :
        """abcd_and_type: abcd_cross_type ('&' abcd_cross_type)*
        -> snakes.typing.Type

        <<< typedef foo : int
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=NamedType(name='int'))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< typedef foo : int & bool
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=IntersectionType(types=[NamedType(name='int'), NamedType(name='bool')]))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        if len(st) == 1 :
            return self.do(st[0])
        else :
            return self.ST.IntersectionType(lineno=st.srow, col_offset=st.scol,
                                            types=[self.do(child) for child in st[::2]])
    def do_abcd_cross_type (self, st, ctx) :
        """abcd_cross_type: abcd_base_type ('*' abcd_base_type)*
        -> snakes.typing.Type

        <<< typedef foo : int
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=NamedType(name='int'))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< typedef foo : int * bool
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=CrossType(types=[NamedType(name='int'), NamedType(name='bool')]))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        if len(st) == 1 :
            return self.do(st[0])
        else :
            return self.ST.CrossType(lineno=st.srow, col_offset=st.scol,
                                     types=[self.do(child) for child in st[::2]])
    def do_abcd_base_type (self, st, ctx) :
        """abcd_base_type: (NAME ['(' abcd_type (',' abcd_type)* ')']
                 | 'enum' '(' test (',' test)* ')' | '(' abcd_type ')')
        -> snakes.typing.Type

        <<< typedef foo : list(int)
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=ListType(items=NamedType(name='int')))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< typedef foo : set(int)
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=SetType(items=NamedType(name='int')))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< typedef foo : dict(int, bool)
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=DictType(keys=NamedType(name='int'), values=NamedType(name='bool')))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< typedef foo : enum(1, 2, 3)
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=EnumType(items=[Num(n=1), Num(n=2), Num(n=3)]))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        <<< typedef foo : int
        ... [True]
        "AbcdSpec(context=[AbcdTypedef(name='foo', type=NamedType(name='int'))], body=AbcdAction(accesses=[], guard=True), asserts=[])"
        """
        if len(st) == 1 :
            return self.ST.NamedType(lineno=st.srow, col_offset=st.scol,
                                     name=st[0].text)
        elif len(st) == 3 :
            return self.do(st[1])
        elif st[0].text in ("list", "set", "tuple") :
            if len(st) > 4 :
                raise ParseError(st.text,
                                 reason="too many arguments for %s type"
                                 % st[0].text)
            if st[0].text == "list" :
                tree = self.ST.ListType
            elif st[0].text == "tuple" :
                tree = self.ST.TupleType
            else :
                tree = self.ST.SetType
            return tree(lineno=st.srow, col_offset=st.scol,
                        items=self.do(st[2]))
        elif st[0].text == "dict" :
            if len(st) > 6 :
                raise ParseError(st.text,
                                 reason="too many arguments for dict type")
            return self.ST.DictType(lineno=st.srow, col_offset=st.scol,
                                    keys=self.do(st[2]),
                                    values=self.do(st[4]))
        elif st[0].text == "enum" :
            return self.ST.EnumType(lineno=st.srow, col_offset=st.scol,
                                    items=[self.do(child) for child in st[2:-1:2]])
        else :
            raise ParseError(st[0].text,
                             reason=("expected 'enum', 'list', 'set' or"
                                     " 'dict' but found '%s'") % st[0].text)
    def do_tfpdef (self, st, ctx) :
        """tfpdef: NAME [':' ('net' | 'buffer' | 'task')]
        -> str, ast.AST?

        <<< net Foo (x, y, n:net, b:buffer, t:task) : [True]
        ... [False]
        "AbcdSpec(context=[AbcdNet(name='Foo', args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None), arg(arg='n', annotation=Name(id='net', ctx=Load())), arg(arg='b', annotation=Name(id='buffer', ctx=Load())), arg(arg='t', annotation=Name(id='task', ctx=Load()))], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=AbcdSpec(context=[], body=AbcdAction(accesses=[], guard=True), asserts=[]))], body=AbcdAction(accesses=[], guard=False), asserts=[])"
        """
        if len(st) == 1 :
            return st[0].text, None
        else :
            return st[0].text, self.ST.Name(lineno=st[2].srow,
                                            col_offset=st[2].scol,
                                            id=st[2].text,
                                            ctx=ctx())

parse = Translator.parse

if __name__ == "__main__" :
    testparser(Translator)



"""
@todo: revise (actually make) documentation
"""

from snakes.plugins import plugin
from snakes.pnml import Tree, loads, dumps
import imp, sys, socket, traceback, operator

class QueryError (Exception) :
    pass

class Query (object) :
    def __init__ (self, name, *larg, **karg) :
        self._name = name
        self._larg = tuple(larg)
        self._karg = dict(karg)
    __pnmltag__ = "query"
    def __pnmldump__ (self) :
        """
        >>> Query('set', 'x', 42).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <query name="set">
          <argument>
           <object type="str">
            x
           </object>
          </argument>
          <argument>
           <object type="int">
            42
           </object>
          </argument>
         </query>
        </pnml>
        >>> Query('test', x=1).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <query name="test">
          <keyword name="x">
           <object type="int">
            1
           </object>
          </keyword>
         </query>
        </pnml>
        >>> Query('test', 'x', 42, y=1).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <query name="test">
          <argument>
           <object type="str">
            x
           </object>
          </argument>
          <argument>
           <object type="int">
            42
           </object>
          </argument>
          <keyword name="y">
           <object type="int">
            1
           </object>
          </keyword>
         </query>
        </pnml>
        >>> Query('set', 'x', Query('call', 'x.upper')).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <query name="set">
          <argument>
           <object type="str">
            x
           </object>
          </argument>
          <argument>
           <query name="call">
            <argument>
             <object type="str">
              x.upper
             </object>
            </argument>
           </query>
          </argument>
         </query>
        </pnml>
        """
        children = []
        for arg in self._larg :
            children.append(Tree("argument", None,
                                 Tree.from_obj(arg)))
        for name, value in self._karg.items() :
            children.append(Tree("keyword", None,
                                 Tree.from_obj(value),
                                 name=name))
        return Tree(self.__pnmltag__, None, *children, **{"name": self._name})
    @classmethod
    def __pnmlload__ (cls, tree) :
        """
        >>> Tree._tag2obj = {'query': Query}
        >>> t = Query('test', 'x', 42, y=1).__pnmldump__()
        >>> q = Query.__pnmlload__(t)
        >>> q._name
        'test'
        >>> q._larg
        ('x', 42)
        >>> q._karg
        {'y': 1}
        """
        larg = (child.child().to_obj()
                for child in tree.get_children("argument"))
        karg = dict((child["name"], child.child().to_obj())
                    for child in tree.get_children("keyword"))
        return cls(tree["name"], *larg, **karg)
    def run (self, envt) :
        """
        >>> import imp
        >>> env = imp.new_module('environment')
        >>> Query('set', 'x', 'hello').run(env)
        >>> env.x
        'hello'
        >>> Query('set', 'x', Query('call', 'x.upper')).run(env)
        >>> env.x
        'HELLO'
        >>> Query('test', 1, 2, 3).run(env)
        Traceback (most recent call last):
         ...
        QueryError: unknown query 'test'
        """
        try :
            handler = getattr(self, "_run_%s" % self._name)
        except AttributeError :
            raise QueryError("unknown query %r" % self._name)
        self._envt = envt
        larg = tuple(a.run(envt) if isinstance(a, self.__class__) else a
                     for a in self._larg)
        karg = dict((n, v.run(envt) if isinstance(v, self.__class__) else v)
                    for n, v in self._karg.items())
        try :
            return handler(*larg, **karg)
        except TypeError :
            cls, val, tb = sys.exc_info()
            try :
                fun, msg = str(val).strip().split("()", 1)
            except :
                raise val
            if fun.startswith("_run_") and hasattr(self, fun) :
                raise TypeError(fun[5:] + "()" + msg)
            raise val
    def _get_object (self, path) :
        obj = self._envt
        for n in path :
            obj = getattr(obj, n)
        return obj
    def _run_set (self, name, value) :
        """
        >>> import imp
        >>> env = imp.new_module('environment')
        >>> Query('set', 'x', 1).run(env)
        >>> env.x
        1
        """
        path = name.split(".")
        setattr(self._get_object(path[:-1]), path[-1], value)
    def _run_get (self, name) :
        """
        >>> import imp
        >>> env = imp.new_module('environment')
        >>> env.x = 2
        >>> Query('get', 'x').run(env)
        2
        """
        path = name.split(".")
        return self._get_object(path)
    def _run_del (self, name) :
        """
        >>> import imp
        >>> env = imp.new_module('environment')
        >>> env.x = 2
        >>> Query('del', 'x').run(env)
        >>> env.x
        Traceback (most recent call last):
         ...
        AttributeError: 'module' object has no attribute 'x'
        """
        path = name.split(".")
        delattr(self._get_object(path[:-1]), path[-1])
    def _run_call (self, fun, *larg, **karg) :
        """
        >>> import imp
        >>> env = imp.new_module('environment')
        >>> env.x = 'hello'
        >>> Query('call', 'x.center', 7).run(env)
        ' hello '
        >>> env.__dict__.update(__builtins__)
        >>> Query('call', Query('call', 'getattr',
        ...                     Query('call', 'x.center', 7),
        ...                     'upper')).run(env)
        ' HELLO '
        """
        if isinstance(fun, str) :
            fun = self._get_object(fun.split("."))
        return fun(*larg, **karg)

@plugin("snakes.nets")
def extend (module) :
    class UDPServer (object) :
        def __init__ (self, port, size=2**20, verbose=0) :
            self._size = size
            self._verbose = verbose
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self._sock.bind(("", port))
            self._env = imp.new_module("snk")
            self._env.__dict__.update(__builtins__)
            self._env.__dict__.update(operator.__dict__)
            self._env.__dict__.update(module.__dict__)
        def recvfrom (self) :
            return self._sock.recvfrom(self._size)
        def sendto (self, data, address) :
            self._sock.sendto(data.strip() + "\n", address)
        def run (self) :
            while True :
                data, address = self.recvfrom()
                data = data.strip()
                if self._verbose :
                    print("# query from %s:%u" % address)
                try :
                    if self._verbose > 1 :
                        print(data)
                    res = loads(data).run(self._env)
                    if res is None :
                        res = Tree("answer", None, status="ok")
                    else :
                        res = Tree("answer", None, Tree.from_obj(res),
                                   status="ok")
                except :
                    cls, val, tb = sys.exc_info()
                    res = Tree("answer", str(val).strip(),
                               error=cls.__name__, status="error")
                    if self._verbose > 1 :
                        print("# error")
                        for entry in traceback.format_exception(cls, val, tb) :
                            for line in entry.splitlines() :
                                print("## %s" % line)
                if self._verbose :
                    if self._verbose > 1 :
                        print("# answer")
                        print(res.to_pnml())
                    elif res["status"] == "error" :
                        print("# answer: %s: %s" % (res["error"], res.data))
                    else :
                        print("# answer: %s" % res["status"])
                self.sendto(res.to_pnml(), address)
    class TCPServer (UDPServer) :
        def __init__ (self, port, size=2**20, verbose=0) :
            self._size = size
            self._verbose = verbose
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._sock.bind(("", port))
            self._sock.listen(1)
            self._env = imp.new_module("snk")
            self._env.__dict__.update(__builtins__)
            self._env.__dict__.update(operator.__dict__)
            self._env.__dict__.update(module.__dict__)
            self._connection = {}
        def recvfrom (self) :
            connection, address = self._sock.accept()
            self._connection[address] = connection
            parts = []
            while True :
                parts.append(connection.recv(self._size))
                if len(parts[-1]) < self._size :
                    break
            return "".join(parts), address
        def sendto (self, data, address) :
            self._connection[address].send(data.rstrip() + "\n")
            self._connection[address].close()
            del self._connection[address]
    return Query, UDPServer, TCPServer



"""Adds methods to draw `PetriNet` and `StateGraph` instances using
GraphViz.

For example, let's first define a Petri net:

>>> import snakes.plugins
>>> snakes.plugins.load('gv', 'snakes.nets', 'nets')
<module ...>
>>> from nets import *
>>> n = PetriNet('N')
>>> n.add_place(Place('p00', [0]))
>>> n.add_transition(Transition('t10'))
>>> n.add_place(Place('p11'))
>>> n.add_transition(Transition('t01'))
>>> n.add_input('p00', 't10', Variable('x'))
>>> n.add_output('p11', 't10', Expression('x+1'))
>>> n.add_input('p11', 't01', Variable('y'))
>>> n.add_output('p00', 't01', Expression('y-1'))

Thanks to plugin `gv`, we can draw it using the various engines of
GraphViz; we can also draw the state graph:

>>> for engine in ('neato', 'dot', 'circo', 'twopi', 'fdp') :
...     n.draw(',test-gv-%s.png' % engine, engine=engine)
Graph(...)
Graph(...)
Graph(...)
Graph(...)
Graph(...)
>>> s = StateGraph(n)
>>> s.build()
>>> s.draw(',test-gv-graph.png')

The plugin also allows to layout the nodes without drawing the net
(this is only available for `PetriNet`, not for `StateGraph`). We
first move every node to position `(-100, -100)`; then, we layout the
net; finally, we check that every node has indeed been moved away from
where we had put it:

>>> for node in n.node() :
...     node.pos.moveto(-100, -100)
>>> all(node.pos.x == node.pos.y == -100 for node in n.node())
True
>>> n.layout()
>>> any(node.pos.x == node.pos.y == -100 for node in n.node())
False

@note: setting nodes position has no influence on how a net is drawn:
    GraphViz will redo the layout in any case. Method `layout` is here
    just in the case you would need a layout of your nets.
@note: this plugin depens on plugins `pos` and `clusters` (that are
    automatically loaded)
"""

import os, os.path, subprocess, collections, codecs
import snakes.plugins
from snakes.plugins.clusters import Cluster
from snakes.compat import *

# apidoc skip
class Graph (Cluster) :
    def __init__ (self, attr) :
        Cluster.__init__(self)
        self.attributes = {}
        self.attr = dict(style="invis")
        self.attr.update(attr)
        self.edges = collections.defaultdict(list)
    def add_node (self, node, attr) :
        self.attributes[node] = attr
        Cluster.add_node(self, node)
    def add_edge (self, src, dst, attr) :
        self.edges[src, dst].append(attr)
    def has_edge (self, src, dst) :
        if (src, dst) in self.edges :
            return True
        for child in self.children() :
            if child.has_edge(src, dst) :
                return True
        return False
    def _dot_attr (self, attr, tag=None) :
        if tag is None :
            tag = ""
        else :
            tag = "%s " % tag
        return (["%s[" % tag,
                 ["%s=%s" % (key, self.escape(unicode(val)))
                  for key, val in attr.items()],
                 "];"])
    def _dot (self) :
        body = []
        lines = ["subgraph %s {" % self, self._dot_attr(self.attr, "graph"),
                 body, "}"]
        for node in sorted(self.nodes()) :
            body.append(node)
            body.append(self._dot_attr(self.attributes[node]))
        for child in self.children() :
            body.extend(child._dot())
        for (src, dst), lst in self.edges.items() :
            for attr in lst :
                body.append("%s -> %s" % (src, dst))
                body.append(self._dot_attr(attr))
        return lines
    def _dot_text (self, lines, indent=0) :
        for l in lines :
            if isinstance(l, (str, unicode)) :
                yield " "*indent*2 + l
            else :
                for x in self._dot_text(l, indent+1) :
                    yield x
    def dot (self) :
        self.done = set()
        return "\n".join(self._dot_text(["digraph {",
#                                         'charset="UTF-8"',
                                         ['node [label="N",'
                                          ' fillcolor="#FFFFFF",'
                                          ' fontcolor="#000000",'
                                          ' style=filled];',
                                          'edge [style="solid"];',
                                          'graph [splines="true",'
                                          ' overlap="false"];'],
                                         self._dot(),
                                         "}"]))
    def escape (self, text) :
        if text.startswith("<") and text.endswith(">") :
            return text
        else :
            return '"%s"' % text.replace('"', r'\"')
    def render (self, filename, engine=None, debug=False) :
        if engine is None :
            engine = getattr(self, "engine", "dot")
        if engine not in ("dot", "neato", "twopi", "circo", "fdp") :
            raise ValueError("unknown GraphViz engine %r" % engine)
        with codecs.open(filename + ".dot", "w", "utf-8") as outfile :
            outfile.write(self.dot())
        if debug :
            dot = subprocess.Popen([engine, "-T" + filename.rsplit(".", 1)[-1],
                                    "-o" + filename, outfile.name],
                                   stdin=subprocess.PIPE,
                                   universal_newlines=True)
        else :
            dot = subprocess.Popen([engine, "-T" + filename.rsplit(".", 1)[-1],
                                    "-o" + filename, outfile.name],
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   universal_newlines=True)
        stdout, stderr = dot.communicate()
        if not debug :
            os.unlink(outfile.name)
        if dot.returncode != 0 :
            if (stdout or "").strip() + (stderr or "").strip() :
                stdout = "\n*** Original error message follows ***\n " + stdout
            raise IOError("%s exited with status %s%s"
                          % (engine, dot.returncode, stdout))
    def layout (self, engine="dot", debug=False) :
        if engine not in ("dot", "neato", "twopi", "circo", "fdp") :
            raise ValueError("unknown GraphViz engine %r" % engine)
        if debug :
            dot = subprocess.Popen([engine, "-Tplain"],
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE)
        else :
            dot = subprocess.Popen([engine, "-Tplain"],
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        if PY3 :
            out, err = dot.communicate(bytes(self.dot(),
                                             snakes.defaultencoding))
            out = out.decode(snakes.defaultencoding)
        else :
            out, err = dot.communicate(self.dot())
        if dot.returncode != 0 :
            raise IOError("%s exited with status %s"
                          % (engine, dot.returncode))
        for line in (l.strip() for l in out.splitlines()
                     if l.strip().startswith("node")) :
            node, name, x, y, rest = line.split(None, 4)
            yield name, float(x), float(y)

@snakes.plugins.plugin("snakes.nets",
                       depends=["snakes.plugins.clusters",
                                "snakes.plugins.pos"])
def extend (module) :
    class PetriNet (module.PetriNet) :
        "An extension with methods `draw` and `layout`"
        def draw (self, filename, engine="dot", debug=False,
                  graph_attr=None, cluster_attr=None,
                  place_attr=None, trans_attr=None, arc_attr=None) :
            """Draw the Petri net to a picture file. How the net is
            rendered can be controlled using the arguments `..._attr`.
            For instance, to draw place in red with names in
            uppercase, and hide True guards, we can proceed as
            follows:

            >>> import snakes.plugins
            >>> snakes.plugins.load('gv', 'snakes.nets', 'nets')
            <module ...>
            >>> from nets import *
            >>> n = PetriNet('N')
            >>> n.add_place(Place('p'))
            >>> n.add_transition(Transition('t'))
            >>> n.add_input('p', 't', Value(dot))
            >>> def draw_place (place, attr) :
            ...     attr['label'] = place.name.upper()
            ...     attr['color'] = '#FF0000'
            >>> def draw_transition (trans, attr) :
            ...     if str(trans.guard) == 'True' :
            ...         attr['label'] = trans.name
            ...     else :
            ...         attr['label'] = '%s\\n%s' % (trans.name, trans.guard)
            >>> n.draw(',net-with-colors.png',
            ...        place_attr=draw_place, trans_attr=draw_transition)
            Graph(...)

            @param filename: the name of the image file to create
            @type filename: `str`
            @param engine: the layout engine to use: `'dot'` (default),
                `'neato'`, `'circo'`, `'twopi'` or `'fdp'`
            @type engine: `str`
            @param place_attr: a function to format places, it will be
                called with the place and its attributes dict as
                parameters
            @type place_attr: `callable`
            @param trans_attr: a function to format transitions, it
                will be called with the transition and its attributes
                dict as parameters
            @type trans_attr: `callable`
            @param arc_attr: a function to format arcs, it will be
                called with the label and its attributes dict as
                parameters
            @type arc_attr: `callable`
            @param cluster_attr: a function to format clusters of
                nodes, it will be called with the cluster and its
                attributes dict as parameters
            @type cluster_attr: `callable`
            """
            nodemap = dict((node.name, "node_%s" % num)
                           for num, node in enumerate(self.node()))
            g = self._copy(nodemap, self.clusters, cluster_attr,
                           place_attr, trans_attr)
            g.engine = engine
            self._copy_edges(nodemap, g, arc_attr)
            if graph_attr :
                graph_attr(self, g.attr)
            if filename is not None :
                g.render(filename, engine, debug)
            g.nodemap = nodemap
            return g
        def _copy (self, nodemap, sub, cluster_attr, place_attr, trans_attr) :
            attr = dict(style="invis")
            if cluster_attr :
                cluster_attr(sub, attr)
            graph = Graph(attr)
            for name in sub.nodes() :
                if self.has_place(name) :
                    node = self.place(name)
                    attr = dict(shape="ellipse",
                                label="%s\\n%s" % (node.name, node.tokens))
                    if place_attr :
                        place_attr(node, attr)
                else :
                    node = self.transition(name)
                    attr = dict(shape="rectangle",
                                label="%s\\n%s" % (node.name, str(node.guard)))
                    if trans_attr :
                        trans_attr(node, attr)
                attr["tooltip"] = node.name
                attr["id"] = nodemap[name]
                graph.add_node(nodemap[name], attr)
            for child in sub.children() :
                graph.add_child(self._copy(nodemap, child, cluster_attr,
                                           place_attr, trans_attr))
            return graph
        def _copy_edges (self, nodemap, graph, arc_attr) :
            for trans in self.transition() :
                for place, label in trans.input() :
                    attr = dict(arrowhead="normal",
                                label=" %s " % label)
                    if arc_attr :
                        arc_attr(label, attr)
                    graph.add_edge(nodemap[place.name], nodemap[trans.name],
                                   attr)
                for place, label in trans.output() :
                    attr = dict(arrowhead="normal",
                                label=" %s " % label)
                    if arc_attr :
                        arc_attr(label, attr)
                    graph.add_edge(nodemap[trans.name], nodemap[place.name],
                                   attr)
        def layout (self, xscale=1.0, yscale=1.0, engine="dot",
                    debug=False, graph_attr=None, cluster_attr=None,
                    place_attr=None, trans_attr=None, arc_attr=None) :
            """Layout the nodes of the Petri net by calling GraphViz
            and reading back the picture it creates. The effect is to
            change attributes `pos` (see plugin `pos`) for every node
            according to the positions calculated by GraphViz.

            @param xscale: how much the image is scaled in the
                horizontal axis after GraphViz has done the layout
            @type xscale: `float`
            @param yscale: how much the image is scaled in the
                vertical axis after GraphViz has done the layout
            @type yscale: `float`
            @param filename: the name of the image file to create
            @type filename: `str`
            @param engine: the layout engine to use: `'dot'` (default),
                `'neato'`, `'circo'`, `'twopi'` or `'fdp'`
            @type engine: `str`
            @param place_attr: a function to format places, it will be
                called with the place and its attributes dict as
                parameters
            @type place_attr: `callable`
            @param trans_attr: a function to format transitions, it
                will be called with the transition and its attributes
                dict as parameters
            @type trans_attr: `callable`
            @param arc_attr: a function to format arcs, it will be
                called with the label and its attributes dict as
                parameters
            @type arc_attr: `callable`
            @param cluster_attr: a function to format clusters of
                nodes, it will be called with the cluster and its
                attributes dict as parameters
            @type cluster_attr: `callable`
            """
            g = self.draw(None, engine, debug, graph_attr, cluster_attr,
                          place_attr, trans_attr, arc_attr)
            node = dict((v, k) for k, v in g.nodemap.items())
            for n, x, y in g.layout(engine, debug) :
                self.node(node[n]).pos.moveto(x*xscale, y*yscale)
    class StateGraph (module.StateGraph) :
        "An extension with a method `draw`"
        def draw (self, filename, engine="dot", debug=False,
                  node_attr=None, edge_attr=None, graph_attr=None) :
            """Draw the state graph to a picture file.

            @param filename: the name of the image file to create or
                `None` if only the computed graph is needed
            @type filename: `None` or `str`
            @param engine: the layout engine to use: 'dot' (default),
                'neato', 'circo', 'twopi' or 'fdp'
            @type engine: `str`
            @param node_attr: a function to format nodes, it will be
                called with the state number, the `StateGraph` object
                and attributes dict as parameters
            @type node_attr: `callable`
            @param edge_attr: a function to format edges, it will be
                called with the transition, its mode and attributes
                dict as parameters
            @type trans_attr:
                `callable`
            @param graph_attr: a function to format grapg, it will be
                called with the state graphe and attributes dict as
                parameters
            @type graph_attr: `callable`
            """
            attr = dict(style="invis",
                        splines="true")
            if graph_attr :
                graph_attr(self, attr)
            graph = Graph(attr)
            for state in self._done :
                self.goto(state)
                attr = dict(shape="rectangle")
                if state == 0 :
                    attr["shape"] = ""
                if node_attr :
                    node_attr(state, self, attr)
                graph.add_node(str(state), attr)
                for succ, trans, mode in self.successors() :
                    attr = dict(arrowhead="normal",
                                label="%s\\n%s" % (trans.name, mode))
                    if edge_attr :
                        edge_attr(trans, mode, attr)
                    graph.add_edge(str(state), str(succ), attr)
            if filename is None :
                return graph
            else :
                graph.render(filename, engine, debug)
    return PetriNet, StateGraph



"""This plugin provides an implementation of the action-based
synchronisation from algebras of Petri nets. With respect to the usual
definition of synchronisation, the plugin is slightly more general in
that it does not impose a fixed arity for action. The extensions are
as follows:

  * class `Action` corresponds to a synchronisable action, it has a
    name, a send/receive flag and a list of parameters. Actions have
    no predetermined arities, only conjugated actions with the same
    arity will be able to synchronise
  * class `MultiAction` corresponds to a multiset of actions. It is
    forbidden to build a multiaction that holds a pair of conjugated
    actions because this leads to infinite nets upon synchronisation
  * the constructor of `Transition` accepts a parameter `actions` that
    is a collection of instances of `Action`, this multiaction is
    added in the attribute `actions` of the transition
  * PetriNet is given new methods: `synchronise` to perform the
    synchronisation, `restrict` to perform the restriction (ie, remove
    transitions with a given action) and `scope` for the scoping (ie,
    synchronisation followed by restriction)

### Example ###

Let's start with an example: we build a Petri net with two transitions
in parallel that will be synchronised later on.

>>> import snakes.plugins
>>> snakes.plugins.load('synchro', 'snakes.nets', 'nets')
<module ...>
>>> from nets import *
>>> n = PetriNet('N')
>>> n.add_place(Place('e1'))
>>> n.add_place(Place('x1'))
>>> m1 = [Action('a', True, [Variable('x'), Value(2)]),
...       Action('b', True, [Value(3), Variable('y')]),
...       Action('c', False, [Variable('x'), Variable('y')])]
>>> print(', '.join(str(action) for action in m1))
a!(x,2), b!(3,y), c?(x,y)
>>> n.add_transition(Transition('t1', guard=Expression('x!=y'), actions=m1))
>>> n.add_input('e1', 't1', Variable('x'))
>>> n.add_output('x1', 't1', Variable('z'))
>>> n.add_place(Place('e2'))
>>> n.add_place(Place('x2'))
>>> m2 = [Action('a', False, [Variable('w'), Variable('y')]),
...       Action('d', False, [Variable('z')])]
>>> print(', '.join(str(action) for action in m2))
a?(w,y), d?(z)
>>> n.add_transition(Transition('t2', guard=Expression('z>0'), actions=m2))
>>> n.add_input('e2', 't2', Variable('w'))
>>> n.add_output('x2', 't2', Variable('z'))
>>> n.transition('t1').vars() == set(['x', 'y', 'z'])
True

On transition `t1`, we have put a multiaction that can be abbreviated
as `a!(x,2), a!(3,y), b?(x,y)` and can be interpreted as three
synchronous communication performed atomically and simultaneously when
`t1` fires:

  * `a!(x,2)` emits `(x,2)` on channel `a`
  * `b!(2,y)` emits `(2,y)` on channel `b`
  * `c?(x,y)` receives `(x,y)` on channel `c`

And similarly for transition `t2` that has two actions:

  * `a?(w,y)` receives `(w,y)` on channel `a`
  * `d?(z)` receives `z` on channel `d`

Thus, `t1` and `t2` hold _conjugated actions_, which are matching
emitting and receiving actions `a!(x,2)` and `a?(x,y)`. So we can
synchronise the net on `a` which builds a new transition whose firing
is exactly equivalent to the simultaneous firing of `t1` and `t2`
performing the communications over channel `a`.

>>> n.synchronise('a')
>>> t = [t for t in n.transition() if t.name not in ('t1', 't2')][0]
>>> print(t.name)
a(w,2)@(t1[x=w,y=e,z=f]+t2[y=2])
>>> print(t.guard)
((w != e)) and ((z > 0))
>>> print(', '.join(sorted(str(action) for action in t.actions)))
b!(3,e), c?(w,e), d?(z)

The second statement `t = ...` retrieves the new transition then we
print its name and guard. The names can be read as: this is the result
of execution action `a(w,2)` on `t1` substituted by `{x->w, y->e,
z->f}` and `t2` substituted by `{y->2}`. Indeed, both transitions have
variables `y` and `z`in common, so they are replaced in `t1` to avoid
names clashes, then, actions a can be `a!(x,2)` and `a?(w,y)` can be
matched by considering `x=w` and `y=2` which yields the rest of the
substitutions for the transitions. The resulting transition results
from the merging of the unified transitions, its guard is the `and` of
the guards of the merged transitions and its multiaction is the union
of the multiactions of the merged transitions minus the actions that
did synchronise.

The net now has three transitions: `t1`, `t2` and the new one
resulting from the synchronisation. This allows both synchronous and
asynchronous behaviour:

>>> for t in sorted(t.name for t in n.transition()) :
...     print(t)
a(w,2)@(t1[x=w,y=e,z=f]+t2[y=2])
t1
t2

If we want to force the synchronous behaviour, we have to restrict
over `'a'` which removes any transition that hold an action `a?(...)`
or `a!(...)`. In practice, this is what we want and so we may have
used method `n.scope('a')` to apply directly the synchronisation
followed by the restriction.

>>> n.restrict('a')
>>> for t in sorted(t.name for t in n.transition()) :
...     print(t)
a(w,2)@(t1[x=w,y=e,z=f]+t2[y=2])
"""

from snakes import ConstraintError
from snakes.data import Substitution, WordSet, iterate
from snakes.nets import Value, Variable
from snakes.pnml import Tree
import snakes.plugins
from snakes.plugins import new_instance
from snakes.compat import *

class Action (object) :
    """Models one action with a name, a direction (send or receive)
    and parameters.
    """
    def __init__ (self, name, send, params) :
        """Constructor. The direction is passed as a Boolean: `True`
        for a send action, `False` for a receive.

        @param name: the name of the action
        @type name: `str`
        @param send: a flag indicating whether this is a send or
            receive action
        @type send: `bool`
        @param params: the list of parameters that must me instances
            of `Variable` or `Value`
        @type params: `list`
        """
        self.name = name
        self.send = send
        self.params = list(params)
    __pnmltag__ = "action"
    # apidoc skip
    def __pnmldump__ (self) :
        """
        >>> Action('a', True, [Value(1), Variable('x')]).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>...
         <action name="a" send="True">
          <value>
           <object type="int">1</object>
          </value>
          <variable>x</variable>
         </action>
        </pnml>
        """
        result = Tree(self.__pnmltag__, None,
                      name=self.name,
                      send=str(self.send))
        for param in self.params :
            result.add_child(Tree.from_obj(param))
        return result
    # apidoc stop
    @classmethod
    def __pnmlload__ (cls, tree) :
        """
        >>> t = Action('a', True, [Value(1), Variable('x')]).__pnmldump__()
        >>> Action.__pnmlload__(t)
        Action('a', True, [Value(1), Variable('x')])
        """
        params = [Tree.to_obj(child) for child in tree.children]
        return cls(tree["name"], tree["send"] == "True", params)
    def __str__ (self) :
        """
        >>> a = Action('a', True, [Value(1), Variable('x')])
        >>> str(a)
        'a!(1,x)'
        >>> a.send = False
        >>> str(a)
        'a?(1,x)'
        """
        if self.send :
            return "%s!(%s)" % (self.name, ",".join([str(p) for p in self]))
        else :
            return "%s?(%s)" % (self.name, ",".join([str(p) for p in self]))
    def __repr__ (self) :
        """
        >>> a = Action('a', True, [Value(1), Variable('x')])
        >>> repr(a)
        "Action('a', True, [Value(1), Variable('x')])"
        >>> a.send = False
        >>> repr(a)
        "Action('a', False, [Value(1), Variable('x')])"
        """
        return "%s(%s, %s, [%s])" % (self.__class__.__name__, repr(self.name),
                                     str(self.send),
                                     ", ".join([repr(p) for p in self]))
    def __len__ (self) :
        """Return the number of parameters, aka the arity of the action.

        >>> len(Action('a', True, [Value(1), Variable('x')]))
        2

        @return: the arity of the action
        @rtype: non negative `int`
        """
        return len(self.params)
    def __iter__ (self) :
        """Iterate on the parameters

        >>> list(Action('a', True, [Value(1), Variable('x')]))
        [Value(1), Variable('x')]
        """
        for action in self.params :
            yield action
    def __eq__ (self, other) :
        """Two actions are equal if they have the same name, same send
        flags and same parameters.

        >>> Action('a', True, [Value(1), Variable('x')]) == Action('a', True, [Value(1), Variable('x')])
        True
        >>> Action('a', True, [Value(1), Variable('x')]) == Action('b', True, [Value(1), Variable('x')])
        False
        >>> Action('a', True, [Value(1), Variable('x')]) == Action('a', False, [Value(1), Variable('x')])
        False
        >>> Action('a', True, [Value(1), Variable('x')]) == Action('a', True, [Value(2), Variable('x')])
        False
        >>> Action('a', True, [Value(1), Variable('x')]) == Action('a', True, [Value(1)])
        False

        @param other: the action to compare
        @type other: `Action`
        @return: `True` if the two actions are equal, `False`
            otherwise
        @rtype: `bool`
        """
        if self.name != other.name :
            return False
        elif self.send != other.send :
            return False
        elif len(self.params) != len(other.params) :
            return False
        for p, q in zip(self.params, other.params) :
            if p != q :
                return False
        return True
    def __ne__ (self, other) :
        return not (self == other)
    def copy (self, subst=None) :
        """Copy the action, optionally substituting its parameters.

        >>> a = Action('a', True, [Variable('x'), Value(2)])
        >>> a.copy()
        Action('a', True, [Variable('x'), Value(2)])
        >>> a = Action('a', True, [Variable('x'), Value(2)])
        >>> a.copy(Substitution(x=Value(3)))
        Action('a', True, [Value(3), Value(2)])

        @param subst: if not `None`, a substitution to apply to the
            parameters of the copy mapping variables names to `Value`
            or `Variable`
        @type subst: `Substitution`
        @return: a copy of the action, substituted by `subst` if not
            `None`
        @rtype: `Action`
        """
        result = self.__class__(self.name, self.send,
                                [p.copy() for p in self.params])
        if subst is not None :
            result.substitute(subst)
        return result
    def substitute (self, subst) :
        """Substitute the parameters according to `subst`

        >>> a = Action('a', True, [Variable('x'), Value(2)])
        >>> a.substitute(Substitution(x=Value(3)))
        >>> a
        Action('a', True, [Value(3), Value(2)])

        @param subst: a substitution to apply to the parameters,
            mapping variables names to `Value` or `Variable`
        @type subst: `Substitution`
        """
        for i, p in enumerate(self.params) :
            if isinstance(p, Variable) and p.name in subst :
                self.params[i] = subst(p.name)
    def vars (self) :
        """
        >>> Action('a', True, [Value(3), Variable('x'), Variable('y'), Variable('x')]).vars() == set(['x', 'y'])
        True

        @return: the set of variable names appearing in the parameters
            of the action
        @rtype: `set` of `str`
        """
        return set(p.name for p in self.params if isinstance(p, Variable))
    def __and__ (self, other) :
        """Compute an unification of two conjugated actions.

        An unification is a `Substitution` that maps variable names to
        `Variable` or `Values`. If both actions are substituted by
        this unification, their parameters lists become equal. If no
        unification can be found, `ConstraintError` is raised (or,
        rarely, `DomainError` depending on the cause of the failure).

        >>> s = Action('a', True, [Value(3), Variable('x'), Variable('y'), Variable('x')])
        >>> r = Action('a', False, [Value(3), Value(2), Variable('t'), Variable('z')])
        >>> u = s & r
        >>> u == Substitution(y=Variable('t'), x=Value(2), z=Value(2))
        True
        >>> s.substitute(u)
        >>> r.substitute(u)
        >>> s.params == r.params
        True
        >>> s.params
        [Value(3), Value(2), Variable('t'), Value(2)]
        >>> s = Action('a', True, [Value(2), Variable('x'), Variable('y'), Variable('x')])
        >>> r = Action('a', False, [Value(3), Value(2), Variable('t'), Variable('z')])
        >>> try : s & r
        ... except ConstraintError : print(sys.exc_info()[1])
        incompatible values
        >>> r = Action('a', False, [Value(2), Value(2), Variable('t')])
        >>> try : s & r
        ... except ConstraintError : print(sys.exc_info()[1])
        arities do not match
        >>> r = Action('b', False, [Value(3), Value(2), Variable('t'), Variable('z')])
        >>> try : s & r
        ... except ConstraintError : print(sys.exc_info()[1])
        actions not conjugated
        >>> r = Action('a', True, [Value(3), Value(2), Variable('t'), Variable('z')])
        >>> try : s & r
        ... except ConstraintError : print(sys.exc_info()[1])
        actions not conjugated

        @param other: the other action to unify with
        @type other: `Action`
        @return: a substitution that unify both actions
        @rtype: `Substitution`
        """
        if (self.name != other.name) or (self.send == other.send) :
            raise ConstraintError("actions not conjugated")
        elif len(self) != len(other) :
            raise ConstraintError("arities do not match")
        result = Substitution()
        for x, y in zip(self.params, other.params) :
            # apply the unifier already computed
            if isinstance(x, Variable) and x.name in result :
                x = result(x.name)
            if isinstance(y, Variable) and y.name in result :
                y = result(y.name)
            # unify the current pair of parameters
            if isinstance(x, Value) and isinstance(y, Value) :
                if x.value != y.value :
                    raise ConstraintError("incompatible values")
            elif isinstance(x, Variable) and isinstance(y, Value) :
                result += Substitution({x.name : y.copy()})
            elif isinstance(x, Value) and isinstance(y, Variable) :
                result += Substitution({y.name : x.copy()})
            elif isinstance(x, Variable) and isinstance(y, Variable) :
                if x.name != y.name :
                    result += Substitution({x.name : y.copy()})
            else :
                raise ConstraintError("unexpected action parameter")
        return result

class MultiAction (object) :
    """Models a multiset of actions.
    """
    def __init__ (self, actions) :
        """The only restriction when building a multiaction is to
        avoid putting two conjugated actions in it. Indeed, this may
        lead to infinite Petri nets upon synchronisation. For example,
        consider two transitions `t1` and `t2` with both `a?()` and
        `a!()` in their multiactions. We can synchronise say `a?()`
        from `t1` with `a!()` from `t2` yielding a transition whose
        multiaction has `a!()` from `t1` and `a?() from `t2` and thus
        can be synchronised with `t1` or `t2`, yielding a new
        transition with `a?()` and `a!()`, etc. Fortunately, it makes
        little sense to let a transition synchronise with itself, so
        this situation is simply forbidden.

        >>> try : MultiAction([Action('a', True, [Variable('x')]),
        ...                    Action('a', False, [Value(2)])])
        ... except ConstraintError : print(sys.exc_info()[1])
        conjugated actions in the same multiaction

        @param actions: a collection of `Action` instances with no
            conjugated actions in it
        @type actions: `iterable`
        """
        self._actions = []
        self._sndrcv = {}
        self._count = {}
        for act in actions :
            self.add(act)
    __pnmltag__ = "multiaction"
    # apidoc stop
    def __pnmldump__ (self) :
        """
        >>> MultiAction([Action('a', True, [Variable('x')]),
        ...              Action('b', False, [Variable('y'), Value(2)])
        ...             ]).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>...
         <multiaction>
          <action name="a" send="True">
           <variable>x</variable>
          </action>
          <action name="b" send="False">
           <variable>y</variable>
           <value>
            <object type="int">2</object>
           </value>
          </action>
         </multiaction>
        </pnml>
        """
        return Tree(self.__pnmltag__, None,
                    *(Tree.from_obj(action) for action in self._actions))
    @classmethod
    def __pnmlload__ (cls, tree) :
        """
        >>> t = MultiAction([Action('a', True, [Variable('x')]),
        ...                  Action('b', False, [Variable('y'), Value(2)])
        ...                 ]).__pnmldump__()
        >>> MultiAction.__pnmlload__(t)
        MultiAction([Action('a', True, [Variable('x')]),
                     Action('b', False, [Variable('y'), Value(2)])])
        """
        return cls(child.to_obj() for child in tree.children)
    def __repr__ (self) :
        """
        >>> MultiAction([Action('a', True, [Variable('x')]),
        ...              Action('b', False, [Variable('y'), Value(2)])])
        MultiAction([Action('a', True, [Variable('x')]),
                     Action('b', False, [Variable('y'), Value(2)])])
        """
        return "%s([%s])" % (self.__class__.__name__,
                             ", ".join(repr(act) for act in self._actions))
    def __str__ (self) :
        """
        >>> str(MultiAction([Action('a', True, [Variable('x')]),
        ...                  Action('b', False, [Variable('y'), Value(2)])]))
        '[a!(x), b?(y,2)]'
        """
        return "[%s]" % ", ".join(str(act) for act in self._actions)
    def send (self, name) :
        """Returns the send flag of the action `name` in this
        multiaction.

        This value is unique as conjugated actions are forbidden in
        the same multiaction.

        >>> m = MultiAction([Action('a', True, [Variable('x')]),
        ...                  Action('b', False, [Variable('y'), Value(2)])])
        >>> m.send('a'), m.send('b')
        (True, False)
        """
        return self._sndrcv[name]
    def add (self, action) :
        """Add an action to the multiaction.

        This may raise `ConstraintError` if the added action is
        conjugated to one that already belongs to the multiaction.

        @param action: the action to add
        @type action: `Action`
        """
        if self._sndrcv.get(action.name, action.send) != action.send :
            raise ConstraintError("conjugated actions in the same multiaction")
        self._sndrcv[action.name] = action.send
        self._count[action.name] = self._count.get(action.name, 0) + 1
        self._actions.append(action)
    def remove (self, action) :
        """Remove an action from the multiaction.

        This may raise `ValueError` if the removed action does belongs
        to the multiaction.

        @param action: the action to remove
        @type action: `Action`
        """
        self._actions.remove(action)
        self._count[action.name] -= 1
        if self._count[action.name] == 0 :
            del self._count[action.name]
            del self._sndrcv[action.name]
    def __iter__ (self) :
        """Iterate over the actions in the multiaction.

        >>> list(MultiAction([Action('a', True, [Variable('x')]),
        ...                   Action('b', False, [Variable('y'), Value(2)])]))
        [Action('a', True, [Variable('x')]),
         Action('b', False, [Variable('y'), Value(2)])]
        """
        for action in self._actions :
            yield action
    def __len__ (self) :
        """Return the number of actions in a multiaction.

        >>> len(MultiAction([Action('a', True, [Variable('x')]),
        ...                  Action('b', False, [Variable('y'), Value(2)])]))
        2

        @return: the number of contained actions
        @rtype: non negative `int`
        """
        return len(self._actions)
    def substitute (self, subst) :
        """Substitute bu `subt` all the actions in the multiaction.

        >>> m = MultiAction([Action('a', True, [Variable('x')]),
        ...                  Action('b', False, [Variable('y'), Variable('x')])])
        >>> m.substitute(Substitution(x=Value(4)))
        >>> m
        MultiAction([Action('a', True, [Value(4)]),
                     Action('b', False, [Variable('y'), Value(4)])])
        """
        for action in self._actions :
            action.substitute(subst)
    def copy (self, subst=None) :
        """Copy the multiaction (and the actions is contains) optionally
        substituting it.

        @param subst: if not `None`, the substitution to apply to the
            copy.
        @type subst: `Substitution`
        @return: a copy of the multiaction, optionally substituted
        @rtype: `MultiAction`
        """
        return self.__class__(act.copy(subst) for act in self._actions)
    def __contains__ (self, action) :
        """Search an action in the multiaction.

        The searched action may be a complete `Action`, just an action
        name, or a pair `(name, send_flag)`.

        >>> m = MultiAction([Action('a', True, [Variable('x'), Value(2)]),
        ...                  Action('a', True, [Value(3), Variable('y')]),
        ...                  Action('b', False, [Variable('x'), Variable('y')])])
        >>> 'a' in m, 'b' in m, 'c' in m
        (True, True, False)
        >>> ('a', True) in m, ('a', False) in m
        (True, False)
        >>> Action('a', True, [Variable('x'), Value(2)]) in m
        True
        >>> Action('a', True, [Variable('x')]) in m
        False
        >>> Action('a', False, [Variable('x'), Value(2)]) in m
        False
        >>> Action('c', True, [Variable('x'), Value(2)]) in m
        False

        @param action: an complete action, or its name or its name and
            send flag
        @type action: `Action`
        @return: `True` if the specified action was found, `False`
            otherwise
        @rtype: `bool`
        """
        if isinstance(action, Action) :
            return action in self._actions
        elif isinstance(action, tuple) and len(action) == 2 :
            return (action[0] in self._sndrcv
                    and self._sndrcv[action[0]] == action[1])
        elif isinstance(action, str) :
            return action in self._count
        else :
            raise ValueError("invalid action specification")
    def __add__ (self, other) :
        """Create a multiaction by adding the actions of two others.

        >>> m = MultiAction([Action('a', True, [Variable('x'), Value(2)]),
        ...                  Action('a', True, [Value(3), Variable('y')]),
        ...                  Action('b', False, [Variable('x'), Variable('y')])])
        >>> m + m
        MultiAction([Action('a', True, [Variable('x'), Value(2)]),
                     Action('a', True, [Value(3), Variable('y')]),
                     Action('b', False, [Variable('x'), Variable('y')]),
                     Action('a', True, [Variable('x'), Value(2)]),
                     Action('a', True, [Value(3), Variable('y')]),
                     Action('b', False, [Variable('x'), Variable('y')])])
        >>> m + Action('c', True, [])
        MultiAction([Action('a', True, [Variable('x'), Value(2)]),
                     Action('a', True, [Value(3), Variable('y')]),
                     Action('b', False, [Variable('x'), Variable('y')]),
                     Action('c', True, [])])

        @param other: the other multiaction to combine or a single
            action
        @type other: `MultiAction` or `Action`
        @return: the concatenated multiaction
        @rtype: `MultiAction`
        """
        if isinstance(other, Action) :
            other = self.__class__([other])
        result = self.copy()
        for action in other._actions :
            result.add(action)
        return result
    def __sub__ (self, other) :
        """Create a multiaction by substracting the actions of two
        others.

        >>> m = MultiAction([Action('a', True, [Variable('x'), Value(2)]),
        ...                  Action('a', True, [Value(3), Variable('y')]),
        ...                  Action('b', False, [Variable('x'), Variable('y')])])
        >>> m - m
        MultiAction([])
        >>> m - Action('b', False, [Variable('x'), Variable('y')])
        MultiAction([Action('a', True, [Variable('x'), Value(2)]),
                     Action('a', True, [Value(3), Variable('y')])])

        @param other: the other multiaction to combine or a single
            action
        @type other: `MultiAction` or `Action`
        @return: the resulting multiaction
        @rtype: `MultiAction`
        """
        if isinstance(other, Action) :
            other = self.__class__([other])
        result = self.copy()
        for action in other._actions :
            result.remove(action)
        return result
    def vars (self) :
        """Return the set of variable names used in all the actions of
        the multiaction.

        >>> MultiAction([Action('a', True, [Variable('x'), Value(2)]),
        ...              Action('a', True, [Value(3), Variable('y')]),
        ...              Action('b', False, [Variable('x'), Variable('z')])]).vars() == set(['x', 'y', 'z'])
        True

        @return: the set of variable names
        @rtype: `set` of `str`
        """
        result = set()
        for action in self._actions :
            result.update(action.vars())
        return result
    def names (self) :
        """Return the set of action names used in the multiaction.

        >>> MultiAction([Action('a', True, [Variable('x'), Value(2)]),
        ...              Action('a', True, [Value(3), Variable('y')]),
        ...              Action('b', False, [Variable('x'), Variable('z')])]).names() == set(['a', 'b'])
        True

        @return: the set of variable names
        @rtype: `set` of `str`
        """
        return set([action.name for action in self._actions])
    def synchronise (self, other, name, common, allnames) :
        """Search all the possible synchronisation on an action name with
        another multiaction.

        This method returns an iterator that yields for each possible
        synchronisation a 4-tuple whose components are:


          * the sending action that did synchronise, it is already
            unified, so the corresponding receiving action is just
            the same with the reversed send flag
          * the multiaction resulting from the synchronisation that
            is also unified
          * the substitution that must be applied to the transition
            that provided the sending action
          * the substitution that must be applied to the transition
            that provided the receiving action

        >>> m = MultiAction([Action('a', True, [Variable('x'), Value(2)]),
        ...                  Action('a', True, [Value(3), Variable('y')]),
        ...                  Action('b', False, [Variable('x'), Variable('y')])])
        >>> n = MultiAction([Action('a', False, [Variable('w'), Variable('y')]),
        ...                  Action('c', False, [Variable('y')])])
        >>> _m, _n = m.vars(), n.vars()
        >>> for a, x, u, v in m.synchronise(n, 'a', _m & _n, _m | _n) :
        ...    print('%s %s' % (str(a), str(x)))
        ...    print(list(sorted(u.items())))
        ...    print(list(sorted(v.items())))
        a!(w,2) [a!(3,d), b?(w,d), c?(2)]
        [('x', Variable('w')), ('y', Variable('d'))]
        [('x', Variable('w')), ('y', Value(2))]
        a!(3,y) [a!(x,2), b?(x,y), c?(y)]
        [('d', Variable('y')), ('w', Value(3)), ('y', Variable('d'))]
        [('d', Variable('y')), ('w', Value(3))]

        @param other: the other multiaction to synchronise with
        @type other: `MultiAction`
        @param name: the name of the action to synchronise on
        @type name: `str`
        @param common: the set of names in common on both transitions
        @type common: `set`
        @param allnames: the set of all names involved in the transitions
        @type allnames: `set`
        @return: an iterator over the possible synchronisations
        @rtype: iterator of `tuple(Action, MultiAction, Substitution,
            Substitution)`
        """
        renamer = Substitution()
        if common :
            names = WordSet(set(allnames) | self.names() | other.names())
            for var in common :
                renamer += Substitution({var : Variable(names.fresh(add=True))})
        for left in (act for act in self._actions if act.name == name) :
            for right in (act for act in other._actions if act.name == name
                          if act.send != left.send) :
                _left = left.copy(renamer)
                try :
                    unifier = _left & right
                except :
                    continue
                _self = self.copy(renamer) - _left
                _self.substitute(unifier)
                _other = other - right
                _other.substitute(unifier)
                yield (_left.copy(unifier), _self + _other,
                       unifier * renamer, unifier)

@snakes.plugins.plugin("snakes.nets")
def extend (module) :
    class Transition (module.Transition) :
        """Class `Transition` is extended to allow a keyword argument
        `actions` in several of its methods `__init__` and `copy` (to
        replace a multiaction upon copy).
        """
        # apidoc stop
        def __init__ (self, name, guard=None, **args) :
            self.actions = MultiAction(args.pop("actions", []))
            module.Transition.__init__(self, name, guard, **args)
        def vars (self) :
            return module.Transition.vars(self) | self.actions.vars()
        def substitute (self, subst) :
            module.Transition.substitute(self, subst)
            self.actions.substitute(subst)
        def copy (self, name=None, **args) :
            actions = args.pop("actions", None)
            result = module.Transition.copy(self, name, **args)
            if actions is None :
                result.actions = self.actions.copy()
            else :
                result.actions = MultiAction(actions)
            return result
        def __pnmldump__ (self) :
            """
            >>> m = MultiAction([Action('a', True, [Variable('x')]),
            ...                  Action('b', False, [Variable('y'), Value(2)])])
            >>> Transition('t', actions=m).__pnmldump__()
            <?xml version="1.0" encoding="utf-8"?>
            <pnml>...
             <transition id="t">
              <multiaction>
               <action name="a" send="True">
                <variable>x</variable>
               </action>
               <action name="b" send="False">
                <variable>y</variable>
                <value>
                 <object type="int">2</object>
                </value>
               </action>
              </multiaction>
             </transition>
            </pnml>
            """
            result = module.Transition.__pnmldump__(self)
            result.add_child(Tree.from_obj(self.actions))
            return result
        @classmethod
        def __pnmlload__ (cls, tree) :
            """
            >>> m = MultiAction([Action('a', True, [Variable('x')]),
            ...                  Action('b', False, [Variable('y'), Value(2)])])
            >>> t = Transition('t', actions=m).__pnmldump__()
            >>> Transition.__pnmlload__(t).actions
            MultiAction([Action('a', True, [Variable('x')]),
                         Action('b', False, [Variable('y'), Value(2)])])
            """
            result = new_instance(cls, module.Transition.__pnmlload__(tree))
            result.actions = Tree.to_obj(tree.child(MultiAction.__pnmltag__))
            return result
    class PetriNet (module.PetriNet) :
        def synchronise (self, name) :
            """Synchronise the net wrt `name`.

            @param name: the action name to be synchronised
            @type name: `str`
            @return: the synchronised Petri net
            @rtype: `PetriNet`
            """
            snd = []
            rcv = []
            for trans in self.transition() :
                if (name, True) in trans.actions :
                    snd.append(trans)
                elif (name, False) in trans.actions :
                    rcv.append(trans)
            loop = True
            done = set()
            while loop :
                loop = False
                for _snd in snd :
                    for _rcv in rcv :
                        if (_snd.name, _rcv.name) in done :
                            continue
                        try :
                            _s, _r = _snd.vars(), _rcv.vars()
                            new = _snd.actions.synchronise(_rcv.actions,
                                                           name,
                                                           _s & _r, _s | _r)
                        except ConstraintError :
                            continue
                        for a, m, s, r in new :
                            t = self._synchronise(
                                _snd, s.restrict(_snd.vars()),
                                _rcv, r.restrict(_rcv.vars()),
                                m, a)
                            if (name, True) in t.actions :
                                snd.append(t)
                                loop = True
                            elif (name, False) in t.actions :
                                rcv.append(t)
                                loop = True
                        done.add((_snd.name, _rcv.name))
        def _synchronise (self, snd, s, rcv, r, actions, sync) :
            def _str (binding) :
                return ",".join("%s=%s" % i for i in sorted(binding.items()))
            collect = []
            for trans, subst in ((snd, s), (rcv, r)) :
                new = "%s[%s]" % (trans.name, _str(subst))
                self.copy_transition(trans.name, new)
                collect.append(new)
                new = self.transition(new)
                for var, val in subst.items() :
                    if isinstance(val, Variable) :
                        new.substitute(Substitution({var : val.name}))
                    for place, label in new.input() :
                        if var in label.vars() :
                            self.remove_input(place.name, new.name)
                            self.add_input(place.name, new.name,
                                           label.replace(Variable(var), val))
                    for place, label in new.output() :
                        if var in label.vars() :
                            self.remove_output(place.name, new.name)
                            self.add_output(place.name, new.name,
                                            label.replace(Variable(var), val))
                new.substitute(subst)
            merged = ("%s@(%s)" %
                      (str(sync).replace("?", "").replace("!", ""),
                       "+".join(collect)))
            self.merge_transitions(merged, collect, actions=actions)
            for name in collect :
                self.remove_transition(name)
            return self.transition(merged)
        def restrict (self, name) :
            """Restrict the net wrt `name`.

            @param name: the action name to be synchronised
            @type name: `str`
            @return: the synchronised Petri net
            @rtype: `PetriNet`
            """
            removed = [trans.name for trans in self.transition()
                       if name in trans.actions]
            for trans in removed :
                self.remove_transition(trans)
        def scope (self,  name) :
            """Scope the net wrt `name`, this is equivalent to apply
            synchronisation followed by restriction on the same
            `name`.

            @param name: the action name to be synchronised
            @type name: `str`
            @return: the synchronised Petri net
            @rtype: `PetriNet`
            """

            self.synchronise(name)
            self.restrict(name)
        def merge_transitions (self, target, sources, **args) :
            """Accepts a keyword parameter `actions` to change the
            multiaction of the resulting transition. If `actions` is
            not given, the multiaction of the new transition is the
            sum of the multiactions of the merged transition.

            @keyword actions: the multiaction of the transition
                resulting from the merge
            @type actions: `MultiAction`
            """
            actions = args.pop("actions", None)
            module.PetriNet.merge_transitions(self, target, sources, **args)
            if actions is None :
                actions = MultiAction()
                for src in sources :
                    actions += self.transition(src).actions
                self.transition(target).actions = actions
            else :
                self.transition(target).actions = MultiAction(actions)
        def copy_transition (self, source, targets, **args) :
            """Accepts a keyword parameter `actions` to change the
            multiaction of the resulting transition. If `actions` is
            not given, the multiaction of the new transition is the
            the same multiaction as the copied transition.

            @keyword actions: the multiaction of the transition
                resulting from the copy
            @type actions: `MultiAction`
            """
            actions = args.pop("actions", None)
            module.PetriNet.copy_transition(self, source, targets, **args)
            if actions is None :
                actions = self.transition(source).actions
            else :
                actions = MultiAction(actions)
            old = self.transition(source)
            for trans in iterate(targets) :
                self.transition(trans).actions = actions.copy()
    return PetriNet, Transition, Action, MultiAction



import sys, os
sys.path.insert(0, ".")
import snakes

def check () :
    VERSION = open("VERSION").read().strip()
    if VERSION == snakes.version :
        sys.exit(0)
    else :
        print("VERSION=%s / snakes.version=%s" % (VERSION, snakes.version))
        sys.exit(1)

def next () :
    old = snakes.version
    ver = [int(v) for v in old.split(".")]
    ver[-1] += 1
    new = ".".join(str(v) for v in ver)
    with open("VERSION", "w") as out :
        out.write("%s\n" % new)
    os.system("sed -i.bak 's/%s/%s/' snakes/__init__.py" % (old, new))

if __name__ == "__main__" :
    if sys.argv[1:] == ["check"] :
        check()
    elif sys.argv[1:] == ["next"] :
        next()
    else :
        print("usage: python version.py [check|next]")
        sys.exit(255)



class Nonce (object) :
    def __init__ (self, agent) :
        self._agent = agent
    def __eq__ (self, other) :
        try :
            return self._agent == other._agent
        except :
            return False
    def __ne__ (self, other) :
        return not self.__eq__(other)
    def __str__ (self) :
        return self.__repr__()
    def __repr__ (self) :
        return "Nonce(%s)" % self._agent

def _cross (sets) :
    if len(sets) == 0 :
        pass
    elif len(sets) == 1 :
        for item in sets[0] :
            yield (item,)
    else :
        for item in sets[0] :
            for others in _cross(sets[1:]) :
                yield (item,) + others


class Spy (object) :
    keywords = set(["crypt", "pub", "priv", "secret", "hash"])
    def __init__ (self, *types) :
        """
        >>> s = Spy(str, int, (str, int, (float, object)))
        >>> s
        <Spy>
        >>> s._subtypes == set([str, int, float, object,
        ...                     (float, object),
        ...                     (str, int, (float, object))])
        True
        """
        self._types = set(types)
        self._subtypes = set()
        todo = set(self._types)
        while len(todo) > 0 :
            t = todo.pop()
            self._subtypes.add(t)
            if isinstance(t, tuple) :
                todo.update(t)
    def __str__ (self) :
        """
        >>> str(Spy(str, int))
        '<Spy>'
        """
        return "<%s>" % self.__class__.__name__
    def __repr__ (self) :
        """
        >>> str(Spy(str, int))
        '<Spy>'
        """
        return self.__str__()
    def __eq__ (self, other) :
        """
        >>> Spy(str, int) == Spy(int, str)
        True
        >>> Spy(str, int) == Spy(int, str, float)
        False
        """
        return self._types == other._types
    def __ne__ (self, other) :
        """
        >>> Spy(str, int) != Spy(int, str, float)
        True
        >>> Spy(str, int) != Spy(int, str)
        False
        """
        return not self.__eq__(other)
    def __hash__ (self) :
        """
        >>> hash(Spy(str, int)) == hash(Spy(int, str))
        True
        >>> hash(Spy(str, int)) == hash(Spy(int, str, float))
        False
        """
        return hash(tuple(sorted(self._types)))
    @classmethod
    def get_type (cls, obj) :
        """
        >>> Spy().get_type(('hello', ('foo', 4), 5))
        (<type 'str'>, (<type 'str'>, <type 'int'>), <type 'int'>)
        """
        t = type(obj)
        if t is tuple :
            if len(obj) > 0 and obj[0] in cls.keywords :
                return (obj[0],) + tuple(cls.get_type(o) for o in obj[1:])
            else :
                return tuple(cls.get_type(o) for o in obj)
        else :
            return t
    @classmethod
    def match (cls, obj, pattern) :
        """
        >>> Spy.match(('hello', 42, (1, 2, 3.4)), (str, int, (1, 2, float)))
        True
        >>> Spy.match(('hello', 42, (1, 2, 3.4)), ('hello', int, (1, 2, float)))
        True
        >>> Spy.match(('hello', 42, (1, 2, 3)), (str, int, (1, 2, float)))
        False
        >>> Spy.match(('hello', 42, (1, 2, 3, 4)), (str, int, (1, 2, float)))
        False
        >>> Spy.match(('hello', 42, (1, 2, 3.4)), ('foo', int, (1, 2, float)))
        False
        """
        if type(obj) == tuple == type(pattern) :
            if len(obj) != len(pattern) :
                return False
            for o, p in zip(obj, pattern) :
                if not cls.match(o, p) :
                    return False
            return True
        elif type(pattern) is type :
            return isinstance(obj, pattern)
        else :
            return obj == pattern
    def message (self, obj) :
        """
        >>> Spy(str, int).message('hello')
        True
        >>> Spy(str, int).message(42)
        True
        >>> Spy(str, int).message((1, 2))
        False
        >>> Spy(str, int).message(3.14)
        False
        """
        return self.get_type(obj) in self._types
    def fragment (self, obj) :
        """
        >>> s = Spy(str, int, (str, int, (float, list)))
        >>> s.fragment('hello')
        True
        >>> s.fragment(3.14)
        True
        >>> s.fragment((3.14, []))
        True
        >>> s.fragment(('hello', 1, (3.14, [])))
        True
        >>> s.fragment((1, 2))
        False
        >>> s.fragment({})
        False
        """
        return self.get_type(obj) in self._subtypes
    def can_decrypt (self, message, knowledge) :
        try :
            if message[0] != "crypt" :
                return False
            key = message[1]
            if key[0] == "pub" :
                return ("priv", key[1]) in knowledge
            elif key[0] == "priv" :
                return ("pub", key[1]) in knowledge
            elif key[0] == "secret" :
                return key in knowledge
        except :
            pass
        return False
    def can_decompose (self, message) :
        try :
            if not isinstance(message, tuple) :
                return False
            elif message[0] not in self.keywords :
                return True
        except :
            pass
        return False
    def learn (self, msg, knowledge) :
        """
        >>> s = Spy((int, int, str), (int, int, (str, str)))
        >>> k = set()
        >>> k = s.learn((1, 2, 'hello'), k)
        >>> for m in sorted(k) :
        ...     print m
        1
        2
        hello
        (1, 1, 'hello')
        (1, 1, ('hash', 'hello'))
        (1, 1, ('hello', 'hello'))
        (1, 2, 'hello')
        (1, 2, ('hash', 'hello'))
        (1, 2, ('hello', 'hello'))
        (2, 1, 'hello')
        (2, 1, ('hash', 'hello'))
        (2, 1, ('hello', 'hello'))
        (2, 2, 'hello')
        (2, 2, ('hash', 'hello'))
        (2, 2, ('hello', 'hello'))
        ('hash', 'hello')
        ('hello', 'hello')
        >>> k = s.learn((2, 3, ('hello', 'world')), k)
        >>> for m in sorted(k) :
        ...     print m
        1
        2
        3
        hello
        world
        (1, 1, 'hello')
        (1, 1, 'world')
        (1, 1, ('hash', 'hello'))
        (1, 1, ('hash', 'world'))
        (1, 1, ('hello', 'hello'))
        (1, 1, ('hello', 'world'))
        (1, 1, ('world', 'hello'))
        (1, 1, ('world', 'world'))
        (1, 2, 'hello')
        (1, 2, 'world')
        (1, 2, ('hash', 'hello'))
        (1, 2, ('hash', 'world'))
        (1, 2, ('hello', 'hello'))
        (1, 2, ('hello', 'world'))
        (1, 2, ('world', 'hello'))
        (1, 2, ('world', 'world'))
        (1, 3, 'hello')
        (1, 3, 'world')
        (1, 3, ('hash', 'hello'))
        (1, 3, ('hash', 'world'))
        (1, 3, ('hello', 'hello'))
        (1, 3, ('hello', 'world'))
        (1, 3, ('world', 'hello'))
        (1, 3, ('world', 'world'))
        (2, 1, 'hello')
        (2, 1, 'world')
        (2, 1, ('hash', 'hello'))
        (2, 1, ('hash', 'world'))
        (2, 1, ('hello', 'hello'))
        (2, 1, ('hello', 'world'))
        (2, 1, ('world', 'hello'))
        (2, 1, ('world', 'world'))
        (2, 2, 'hello')
        (2, 2, 'world')
        (2, 2, ('hash', 'hello'))
        (2, 2, ('hash', 'world'))
        (2, 2, ('hello', 'hello'))
        (2, 2, ('hello', 'world'))
        (2, 2, ('world', 'hello'))
        (2, 2, ('world', 'world'))
        (2, 3, 'hello')
        (2, 3, 'world')
        (2, 3, ('hash', 'hello'))
        (2, 3, ('hash', 'world'))
        (2, 3, ('hello', 'hello'))
        (2, 3, ('hello', 'world'))
        (2, 3, ('world', 'hello'))
        (2, 3, ('world', 'world'))
        (3, 1, 'hello')
        (3, 1, 'world')
        (3, 1, ('hash', 'hello'))
        (3, 1, ('hash', 'world'))
        (3, 1, ('hello', 'hello'))
        (3, 1, ('hello', 'world'))
        (3, 1, ('world', 'hello'))
        (3, 1, ('world', 'world'))
        (3, 2, 'hello')
        (3, 2, 'world')
        (3, 2, ('hash', 'hello'))
        (3, 2, ('hash', 'world'))
        (3, 2, ('hello', 'hello'))
        (3, 2, ('hello', 'world'))
        (3, 2, ('world', 'hello'))
        (3, 2, ('world', 'world'))
        (3, 3, 'hello')
        (3, 3, 'world')
        (3, 3, ('hash', 'hello'))
        (3, 3, ('hash', 'world'))
        (3, 3, ('hello', 'hello'))
        (3, 3, ('hello', 'world'))
        (3, 3, ('world', 'hello'))
        (3, 3, ('world', 'world'))
        ('hash', 'hello')
        ('hash', 'world')
        ('hello', 'hello')
        ('hello', 'world')
        ('world', 'hello')
        ('world', 'world')

        >>> s = Spy(('crypt', ('pub', int), str))
        >>> pub, priv = ('pub', 1), ('priv', 1)
        >>> k = set([pub])
        >>> k = s.learn(('crypt', priv, 'hello'), k)
        >>> 'hello' in k
        True
        >>> k = s.learn(('crypt', pub, 'secret message'), k)
        >>> 'secret message' in k
        False
        >>> k.add(priv)
        >>> k = s.learn(('crypt', pub, 'secret message'), k)
        >>> 'secret message' in k
        True
        """
        k = set(knowledge)
        # learn from new message
        # add new message to knowledge
        k.add(msg)
        # hash new message if useful
        h = ("hash", msg)
        if self.fragment(h) :
            k.add(h)
        # try to decrypt new message
        if self.can_decrypt(msg, k) :
            for m in msg[2:] :
                if m not in k :
                    k.update(self.learn(m, k))
        # try to decompose new message
        elif self.can_decompose(msg) :
            for m in msg :
                if m not in k :
                    k.update(self.learn(m, k))
        self._learn_(msg, k)
        # compose new messages from fragments
        for sub in (s for s in sorted(self._subtypes, key=self._size)
                    if type(s) is tuple) :
            sets = []
            for t in sub :
                sets.append([x for x in k|self.keywords if self.match(x, t)])
            k.update(_cross(sets))
        return k
    @classmethod
    def _size (cls, obj) :
        if isinstance(obj, tuple) :
            return (len(obj),) + tuple(cls._size(o) for o in obj)
        else :
            return 1
    def _learn_ (self, m, k) :
        for attr in (a for a in dir(self) if a.startswith("learn_")) :
            getattr(self, attr)(m, k)

class SpyKS (Spy) :
    def can_decrypt (self, message, knowledge) :
        try :
            if message[1][0] == "priv" :
                knowledge.add(("pub", message[1][1]))
        except :
            pass
        return Spy.can_decrypt(self, message, knowledge)

if __name__ == "__main__" :
    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""A plugin to record for each not to which modules it belongs.

Modules are arbitrary values, each `PetriNet`, `Place` or transition
is attached a set of modules. These modules can be given at
construction time, as a single value or a collection of values.
Moreover, when nodes are attached to a net, they inherit the modules
from this net. When nodes are merged, their sets of modules is added
to build the set of modules of the new node.

The plugin implements a very basic notion of modules, un particular,
it does not support hierarchy and when a module is attached to a node,
it is not possible to remove it. So modules are only a way to remember
where a node comes from and to tag nodes as we aggregate nets, e.g.,
through compositions (see plugin `ops`).

Internally, modules are stored in a label called `"modules"` (see
plugin `snakes.plugins.labels`).

>>> import snakes.plugins
>>> snakes.plugins.load(["ops", "modules"], "snakes.nets", "snk")
<module 'snk' ...>
>>> from snk import *
>>> n1 = PetriNet("n1", modules="hello")
>>> n1.add_place(Place("p"))
>>> n1.place("p").modules()
set(['hello'])
>>> n1.add_place(Place("q"))
>>> n2 = PetriNet("n2", modules="world")
>>> n2.add_place(Place("r"))
>>> n2.add_place(Place("s", modules="spam"))
>>> n = n1|n2
>>> list(sorted((p.name, list(sorted(p.modules()))) for p in n.place()))
[('[p|]', ['hello']),
 ('[q|]', ['hello']),
 ('[|r]', ['world']),
 ('[|s]', ['spam', 'world'])]
>>> n.modules("egg")
>>> list(sorted((p.name, list(sorted(p.modules()))) for p in n.place()))
[('[p|]', ['egg', 'hello']),
 ('[q|]', ['egg', 'hello']),
 ('[|r]', ['egg', 'world']),
 ('[|s]', ['egg', 'spam', 'world'])]
"""

from snakes.plugins import plugin, new_instance
from snakes.data import iterate
from snakes.pnml import Tree

@plugin("snakes.nets",
        depends=["snakes.plugins.labels"])
def extend (module) :
    class Transition (module.Transition) :
        def __init__ (self, name, guard=None, **options) :
            mod = set(iterate(options.pop("modules", [])))
            module.Transition.__init__(self, name, guard, **options)
            self.modules(mod)
        def modules (self, modules=None) :
            if modules is None :
                return self.label("modules")
            else :
                self.label(modules=set(iterate(modules)))
    class Place (module.Place) :
        def __init__ (self, name, tokens=[], check=None, **options) :
            mod = set(iterate(options.pop("modules", [])))
            module.Place.__init__(self, name, tokens, check, **options)
            self.modules(mod)
        def modules (self, modules=None) :
            if modules is None :
                return self.label("modules")
            else :
                self.label(modules=set(iterate(modules)))
    class PetriNet (module.PetriNet) :
        def __init__ (self, name, **options) :
            mod = set(iterate(options.pop("modules", [])))
            module.PetriNet.__init__(self, name, **options)
            self.modules(mod)
        def modules (self, modules=None) :
            if modules is None :
                return self.label("modules")
            mod = set(iterate(modules))
            self.label(modules=mod)
            for node in self.node() :
                node.modules(mod | node.modules())
        def add_place (self, place, **options) :
            mod = set(iterate(options.pop("modules", self.modules())))
            module.PetriNet.add_place(self, place, **options)
            place.modules(place.modules() | mod)
        def add_transition (self, trans, **options) :
            mod = set(iterate(options.pop("modules", self.modules())))
            module.PetriNet.add_transition(self, trans, **options)
            trans.modules(trans.modules() | mod)
        def merge_places (self, target, sources, **options) :
            mod = set(iterate(options.pop("modules", self.modules())))
            module.PetriNet.merge_places(self, target, sources, **options)
            new = self.place(target)
            new.modules(reduce(set.__or__,
                               (self.place(p).modules()
                                for p in sources),
                               mod))
        def merge_transitions (self, target, sources, **options) :
            mod = set(iterate(options.pop("modules", self.modules())))
            module.PetriNet.merge_transitions(self, target, sources, **options)
            new = self.transition(target)
            new.modules(reduce(set.__or__,
                               (self.place(p).modules()
                                for p in sources),
                               mod))
    return Transition, Place, PetriNet



"""Python 2 and 3 compatibility layer
"""

import sys

try :
    xrange
except NameError :
    xrange = range

try :
    reduce
except NameError :
    from functools import reduce

try :
    import StringIO as io
except ImportError :
    import io

try :
    next
except NameError :
    def next (obj) :
        return obj.next()

try :
    unicode
except NameError :
    unicode = str

try :
    reduce
except NameError :
    from functools import reduce

try :
    unicode
except NameError :
    unicode = str

PY3 = sys.version > "3"
